import unittest
from unittest import mock

from aprsd.client.drivers.fake import APRSDFakeDriver
from aprsd.packets import core


class TestAPRSDFakeDriver(unittest.TestCase):
    """Unit tests for the APRSDFakeDriver class."""

    def setUp(self):
        # Mock CONF for testing
        self.conf_patcher = mock.patch('aprsd.client.drivers.fake.CONF')
        self.mock_conf = self.conf_patcher.start()

        # Configure fake_client.enabled
        self.mock_conf.fake_client.enabled = True

        # Create an instance of the driver
        self.driver = APRSDFakeDriver()

    def tearDown(self):
        self.conf_patcher.stop()

    def test_init(self):
        """Test initialization sets default values."""
        self.assertEqual(self.driver.path, ['WIDE1-1', 'WIDE2-1'])
        self.assertFalse(self.driver.thread_stop)

    def test_is_enabled_true(self):
        """Test is_enabled returns True when configured."""
        self.mock_conf.fake_client.enabled = True
        self.assertTrue(APRSDFakeDriver.is_enabled())

    def test_is_enabled_false(self):
        """Test is_enabled returns False when not configured."""
        self.mock_conf.fake_client.enabled = False
        self.assertFalse(APRSDFakeDriver.is_enabled())

    def test_is_alive(self):
        """Test is_alive returns True when thread_stop is False."""
        self.driver.thread_stop = False
        self.assertTrue(self.driver.is_alive())

        self.driver.thread_stop = True
        self.assertFalse(self.driver.is_alive())

    def test_close(self):
        """Test close sets thread_stop to True."""
        self.driver.thread_stop = False
        self.driver.close()
        self.assertTrue(self.driver.thread_stop)

    @mock.patch('aprsd.client.drivers.fake.LOG')
    def test_setup_connection(self, mock_log):
        """Test setup_connection does nothing (it's fake)."""
        self.driver.setup_connection()
        # Method doesn't do anything, so just verify it doesn't crash

    def test_set_filter(self):
        """Test set_filter method does nothing (it's fake)."""
        # Just test it doesn't fail
        self.driver.set_filter('test/filter')

    def test_login_success(self):
        """Test login_success always returns True."""
        self.assertTrue(self.driver.login_success())

    def test_login_failure(self):
        """Test login_failure always returns None."""
        self.assertIsNone(self.driver.login_failure())

    @mock.patch('aprsd.client.drivers.fake.LOG')
    def test_send_with_packet_object(self, mock_log):
        """Test send with a Packet object."""
        mock_packet = mock.MagicMock(spec=core.Packet)
        mock_packet.payload = 'Test payload'
        mock_packet.to_call = 'TEST'
        mock_packet.from_call = 'FAKE'

        self.driver.send(mock_packet)

        mock_log.info.assert_called_once()
        mock_packet.prepare.assert_called_once()

    @mock.patch('aprsd.client.drivers.fake.LOG')
    def test_send_with_non_packet_object(self, mock_log):
        """Test send with a non-Packet object."""
        # Create a mock message-like object
        mock_msg = mock.MagicMock()
        mock_msg.raw = 'Test'
        mock_msg.msgNo = '123'
        mock_msg.to_call = 'TEST'
        mock_msg.from_call = 'FAKE'

        self.driver.send(mock_msg)

        mock_log.info.assert_called_once()
        mock_log.debug.assert_called_once()

    @mock.patch('aprsd.client.drivers.fake.LOG')
    @mock.patch('aprsd.client.drivers.fake.time.sleep')
    def test_consumer_with_raw_true(self, mock_sleep, mock_log):
        """Test consumer with raw=True."""
        mock_callback = mock.MagicMock()

        self.driver.consumer(mock_callback, raw=True)

        # Verify callback was called with raw data
        mock_callback.assert_called_once()
        call_args = mock_callback.call_args[1]
        self.assertIn('raw', call_args)
        mock_sleep.assert_called_once_with(1)

    @mock.patch('aprsd.client.drivers.fake.LOG')
    @mock.patch('aprsd.client.drivers.fake.aprslib.parse')
    @mock.patch('aprsd.client.drivers.fake.core.factory')
    @mock.patch('aprsd.client.drivers.fake.time.sleep')
    def test_consumer_with_raw_false(
        self, mock_sleep, mock_factory, mock_parse, mock_log
    ):
        """Test consumer with raw=False."""
        mock_callback = mock.MagicMock()
        mock_packet = mock.MagicMock(spec=core.Packet)
        mock_factory.return_value = mock_packet

        self.driver.consumer(mock_callback, raw=False)

        # Verify the packet was created and passed to callback
        mock_parse.assert_called_once()
        mock_factory.assert_called_once()
        mock_callback.assert_called_once_with(packet=mock_packet)
        mock_sleep.assert_called_once_with(1)

    def test_consumer_updates_keepalive(self):
        """Test consumer updates keepalive timestamp."""
        mock_callback = mock.MagicMock()
        old_keepalive = self.driver.aprsd_keepalive

        # Force a small delay to ensure timestamp changes
        import time

        time.sleep(0.01)

        with mock.patch('aprsd.client.drivers.fake.time.sleep'):
            self.driver.consumer(mock_callback)

        self.assertNotEqual(old_keepalive, self.driver.aprsd_keepalive)
        self.assertGreater(self.driver.aprsd_keepalive, old_keepalive)

    def test_decode_packet_with_empty_kwargs(self):
        """Test decode_packet with empty kwargs."""
        result = self.driver.decode_packet()
        self.assertIsNone(result)

    def test_decode_packet_with_packet(self):
        """Test decode_packet with packet in kwargs."""
        mock_packet = mock.MagicMock(spec=core.Packet)
        result = self.driver.decode_packet(packet=mock_packet)
        self.assertEqual(result, mock_packet)

    @mock.patch('aprsd.client.drivers.fake.aprslib.parse')
    @mock.patch('aprsd.client.drivers.fake.core.factory')
    def test_decode_packet_with_raw(self, mock_factory, mock_parse):
        """Test decode_packet with raw in kwargs."""
        mock_packet = mock.MagicMock(spec=core.Packet)
        mock_factory.return_value = mock_packet
        raw_data = 'raw packet data'

        result = self.driver.decode_packet(raw=raw_data)

        mock_parse.assert_called_once_with(raw_data)
        mock_factory.assert_called_once_with(mock_parse.return_value)
        self.assertEqual(result, mock_packet)

    def test_stats(self):
        """Test stats returns correct information."""
        self.driver.thread_stop = False
        result = self.driver.stats()

        self.assertEqual(result['driver'], 'APRSDFakeDriver')
        self.assertTrue(result['is_alive'])

        # Test with serializable parameter
        result_serializable = self.driver.stats(serializable=True)
        self.assertEqual(result_serializable['driver'], 'APRSDFakeDriver')
        self.assertTrue(result_serializable['is_alive'])


if __name__ == '__main__':
    unittest.main()



import logging
import typing as t
from functools import update_wrapper
from pathlib import Path

import click
from oslo_config import cfg

import aprsd
from aprsd import conf  # noqa: F401
from aprsd.log import log
from aprsd.utils import trace

CONF = cfg.CONF
home = str(Path.home())
DEFAULT_CONFIG_DIR = f'{home}/.config/aprsd/'
DEFAULT_SAVE_FILE = f'{home}/.config/aprsd/aprsd.p'
DEFAULT_CONFIG_FILE = f'{home}/.config/aprsd/aprsd.conf'


F = t.TypeVar('F', bound=t.Callable[..., t.Any])

common_options = [
    click.option(
        '--loglevel',
        default='INFO',
        show_default=True,
        type=click.Choice(
            ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'],
            case_sensitive=False,
        ),
        show_choices=True,
        help='The log level to use for aprsd.log',
    ),
    click.option(
        '-c',
        '--config',
        'config_file',
        show_default=True,
        default=DEFAULT_CONFIG_FILE,
        help='The aprsd config file to use for options.',
    ),
    click.option(
        '--quiet',
        is_flag=True,
        default=False,
        help="Don't log to stdout",
    ),
]


class AliasedGroup(click.Group):
    def command(self, *args, **kwargs):
        """A shortcut decorator for declaring and attaching a command to
        the group.  This takes the same arguments as :func:`command` but
        immediately registers the created command with this instance by
        calling into :meth:`add_command`.
        Copied from `click` and extended for `aliases`.
        """

        def decorator(f):
            aliases = kwargs.pop('aliases', [])
            cmd = click.decorators.command(*args, **kwargs)(f)
            self.add_command(cmd)
            for alias in aliases:
                self.add_command(cmd, name=alias)
            return cmd

        return decorator

    def group(self, *args, **kwargs):
        """A shortcut decorator for declaring and attaching a group to
        the group.  This takes the same arguments as :func:`group` but
        immediately registers the created command with this instance by
        calling into :meth:`add_command`.
        Copied from `click` and extended for `aliases`.
        """

        def decorator(f):
            aliases = kwargs.pop('aliases', [])
            cmd = click.decorators.group(*args, **kwargs)(f)
            self.add_command(cmd)
            for alias in aliases:
                self.add_command(cmd, name=alias)
            return cmd

        return decorator


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def process_standard_options(f: F) -> F:
    def new_func(*args, **kwargs):
        ctx = args[0]
        ctx.ensure_object(dict)
        config_file_found = True
        if kwargs['config_file']:
            default_config_files = [kwargs['config_file']]
        else:
            default_config_files = None

        try:
            CONF(
                [],
                project='aprsd',
                version=aprsd.__version__,
                default_config_files=default_config_files,
            )
        except cfg.ConfigFilesNotFoundError:
            config_file_found = False
        ctx.obj['loglevel'] = kwargs['loglevel']
        # ctx.obj["config_file"] = kwargs["config_file"]
        ctx.obj['quiet'] = kwargs['quiet']
        log.setup_logging(
            ctx.obj['loglevel'],
            ctx.obj['quiet'],
        )
        if CONF.trace_enabled:
            trace.setup_tracing(['method', 'api'])

        if not config_file_found:
            LOG = logging.getLogger('APRSD')  # noqa: N806
            LOG.error("No config file found!! run 'aprsd sample-config'")

        del kwargs['loglevel']
        del kwargs['config_file']
        del kwargs['quiet']
        return f(*args, **kwargs)

    return update_wrapper(t.cast(F, new_func), f)


def process_standard_options_no_config(f: F) -> F:
    """Use this as a decorator when config isn't needed."""

    def new_func(*args, **kwargs):
        ctx = args[0]
        ctx.ensure_object(dict)
        ctx.obj['loglevel'] = kwargs['loglevel']
        ctx.obj['config_file'] = kwargs['config_file']
        ctx.obj['quiet'] = kwargs['quiet']
        log.setup_logging(
            ctx.obj['loglevel'],
            ctx.obj['quiet'],
        )

        del kwargs['loglevel']
        del kwargs['config_file']
        del kwargs['quiet']
        return f(*args, **kwargs)

    return update_wrapper(t.cast(F, new_func), f)



#
# @author Sinu John
#         sinuvian at gmail dot com
#         www.sinujohn.wordpress.com
#

#
# Usage:
#       python clock.py [degree [time]]
#
#           'degree' is the degree of fuzziness - can have values of 1 or 2
#           'time' - give time in the format hour:minute (24-hour format)
#                               example: 9:45, 11:30, 14:25
#            Default value of 'degree' is 1, and 'time' is System time.
#

import sys
import time


def fuzzy(hour, minute, degree=1):
    """Implements the fuzzy clock.
    returns the the string that spells out the time - hour:minute
    Supports two degrees of fuzziness. Set with degree = 1 or degree = 2
    When degree = 1, time is in quantum of 5 minutes.
    When degree = 2, time is in quantum of 15 minutes."""

    if degree <= 0 or degree > 2:
        print("Please use a degree of 1 or 2. Using fuzziness degree=1")
        degree = 1

    begin = "It's "

    f0 = "almost "
    f1 = "exactly "
    f2 = "around "

    b0 = " past "
    b1 = " to "

    hourlist = (
        "One",
        "Two",
        "Three",
        "Four",
        "Five",
        "Six",
        "Seven",
        "Eight",
        "Nine",
        "Ten",
        "Eleven",
        "Twelve",
    )

    s1 = s2 = s3 = s4 = ""
    base = 5

    if degree == 1:
        base = 5
        val = ("Five", "Ten", "Quarter", "Twenty", "Twenty-Five", "Half")
    elif degree == 2:
        base = 15
        val = ("Quarter", "Half")

    # to find whether we have to use 'almost', 'exactly' or 'around'
    dmin = minute % base
    if minute > 30:
        pos = int((60 - minute) / base)  # position in the tuple 'val'
    else:
        pos = int(minute / base)

    if dmin == 0:
        s1 = f1
        pos = pos - 1
    elif dmin <= base / 2:
        s1 = f2
        if minute < 30:
            pos = pos - 1
    else:
        s1 = f0
        if minute > 30:
            pos = pos - 1

    s2 = val[pos]

    if minute <= base / 2:
        # Case like "It's around/exactly Ten"
        s2 = s3 = ""
        s4 = hourlist[hour - 12 - 1]
    elif minute >= 60 - base / 2:
        # Case like "It's almost Ten"
        s2 = s3 = ""
        s4 = hourlist[hour - 12]
    else:
        # Other cases with all words, like "It's around Quarter past One"
        if minute > 30:
            s3 = b1  # to
            s4 = hourlist[hour - 12]
        else:
            s3 = b0  # past
            s4 = hourlist[hour - 12 - 1]

    return begin + s1 + s2 + s3 + s4


def main():
    deg = 1
    stm = time.localtime()
    h = stm.tm_hour
    m = stm.tm_min

    if len(sys.argv) >= 2:
        try:
            deg = int(sys.argv[1])
        except Exception:
            print("Please use a degree of 1 or 2. Using fuzziness degree=1")

        if len(sys.argv) >= 3:
            tm = sys.argv[2].split(":")
            try:
                h = int(tm[0])
                m = int(tm[1])
                if h < 0 or h > 23 or m < 0 or m > 59:
                    raise Exception
            except Exception:
                print("Bad time entered. Using the system time.")
                h = stm.tm_hour
                m = stm.tm_min
                print(fuzzy(h, m, deg))
            return


if __name__ == "__main__":
    main()



"""
The options for log setup
"""

import logging

from oslo_config import cfg

LOG_LEVELS = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
}

DEFAULT_DATE_FORMAT = '%m/%d/%Y %I:%M:%S %p'
DEFAULT_LOG_FORMAT = (
    '[%(asctime)s] [%(threadName)-20.20s] [%(levelname)-5.5s]'
    ' %(message)s - [%(pathname)s:%(lineno)d]'
)

DEFAULT_LOG_FORMAT = (
    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
    '<yellow>{thread.name: <18}</yellow> | '
    '<level>{level: <8}</level> | '
    '<level>{message}</level> | '
    '<cyan>{name}</cyan>:<cyan>{function:}</cyan>:<magenta>{line:}</magenta>'
)

logging_group = cfg.OptGroup(
    name='logging',
    title='Logging options',
)
logging_opts = [
    cfg.StrOpt(
        'logfile',
        default=None,
        help='File to log to',
    ),
    cfg.StrOpt(
        'logformat',
        default=DEFAULT_LOG_FORMAT,
        help='Log file format, unless rich_logging enabled.',
    ),
    cfg.StrOpt(
        'log_level',
        default='INFO',
        choices=LOG_LEVELS.keys(),
        help='Log level for logging of events.',
    ),
    cfg.BoolOpt(
        'enable_color',
        default=True,
        help='Enable ANSI color codes in logging',
    ),
    cfg.BoolOpt(
        'enable_console_stdout',
        default=True,
        help='Enable logging to the console/stdout.',
    ),
]


def register_opts(config):
    config.register_group(logging_group)
    config.register_opts(logging_opts, group=logging_group)


def list_opts():
    return {
        logging_group.name: (logging_opts),
    }



from typing import Callable, Protocol, runtime_checkable

from aprsd.packets import core
from aprsd.utils import singleton, trace


@runtime_checkable
class ClientDriver(Protocol):
    """Protocol for APRSD client drivers.

    This protocol defines the methods that must be
    implemented by APRSD client drivers.
    """

    @staticmethod
    def is_enabled(self) -> bool:
        pass

    @staticmethod
    def is_configured(self) -> bool:
        pass

    @property
    def is_alive(self) -> bool:
        pass

    def close(self) -> None:
        pass

    def send(self, packet: core.Packet) -> bool:
        pass

    def setup_connection(self) -> None:
        pass

    def set_filter(self, filter: str) -> None:
        pass

    def login_success(self) -> bool:
        pass

    def login_failure(self) -> str:
        pass

    def consumer(self, callback: Callable, raw: bool = False) -> None:
        pass

    def decode_packet(self, *args, **kwargs) -> core.Packet:
        pass

    def stats(self, serializable: bool = False) -> dict:
        pass


@singleton
class DriverRegistry(metaclass=trace.TraceWrapperMetaclass):
    """Registry for APRSD client drivers.

    This registry is used to register and unregister APRSD client drivers.

    This allows us to dynamically load the configured driver at runtime.

    All drivers are registered, then when aprsd needs the client, the
    registry provides the configured driver for the single instance of the
    single APRSD client.
    """

    def __init__(self):
        self.drivers = []

    def register(self, driver: Callable):
        if not isinstance(driver, ClientDriver):
            raise ValueError('Driver must be of ClientDriver type')
        self.drivers.append(driver)

    def unregister(self, driver: Callable):
        if driver in self.drivers:
            self.drivers.remove(driver)
        else:
            raise ValueError(f'Driver {driver} not found')

    def get_driver(self) -> ClientDriver:
        """Get the first enabled driver."""
        for driver in self.drivers:
            if driver.is_enabled() and driver.is_configured():
                return driver()
        raise ValueError('No enabled driver found')



import logging

import aprsd
from aprsd import plugin
from aprsd.stats import collector


LOG = logging.getLogger("APRSD")


class VersionPlugin(plugin.APRSDRegexCommandPluginBase):
    """Version of APRSD Plugin."""

    command_regex = r"^([v]|[v]\s|version)"
    command_name = "version"
    short_description = "What is the APRSD Version"

    # message_number:time combos so we don't resend the same email in
    # five mins {int:int}
    email_sent_dict = {}

    def process(self, packet):
        LOG.info("Version COMMAND")
        # fromcall = packet.get("from")
        # message = packet.get("message_text", None)
        # ack = packet.get("msgNo", "0")
        s = collector.Collector().collect()
        return "APRSD ver:{} uptime:{}".format(
            aprsd.__version__,
            s["APRSDStats"]["uptime"],
        )



# Fetch active stats from a remote running instance of aprsd admin web interface.
import logging
import sys

import click
import requests
from oslo_config import cfg
from rich.console import Console
from rich.table import Table

# local imports here
import aprsd
from aprsd import cli_helper
from aprsd.main import cli
from aprsd.threads.stats import StatsStore

# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
LOG = logging.getLogger('APRSD')
CONF = cfg.CONF


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    '--host',
    type=str,
    default=None,
    help='IP address of the remote aprsd admin web ui fetch stats from.',
)
@click.option(
    '--port',
    type=int,
    default=None,
    help='Port of the remote aprsd web admin interface to fetch stats from.',
)
@click.pass_context
@cli_helper.process_standard_options
def fetch_stats(ctx, host, port):
    """Fetch stats from a APRSD admin web interface."""
    console = Console()
    console.print(f'Python version: {sys.version}')
    console.print(f'APRSD Fetch-Stats started version: {aprsd.__version__}')

    CONF.log_opt_values(LOG, logging.DEBUG)
    if not host:
        host = CONF.admin.web_ip
    if not port:
        port = CONF.admin.web_port

    msg = f'Fetching stats from {host}:{port}'
    console.print(msg)
    with console.status(msg):
        response = requests.get(f'http://{host}:{port}/stats', timeout=120)
        if not response:
            console.print(
                f'Failed to fetch stats from {host}:{port}?',
                style='bold red',
            )
            return

        stats = response.json()
        if not stats:
            console.print(
                f'Failed to fetch stats from aprsd admin ui at {host}:{port}',
                style='bold red',
            )
            return

    aprsd_title = (
        'APRSD '
        f'[bold cyan]v{stats["APRSDStats"]["version"]}[/] '
        f'Callsign [bold green]{stats["APRSDStats"]["callsign"]}[/] '
        f'Uptime [bold yellow]{stats["APRSDStats"]["uptime"]}[/]'
    )

    console.rule(f'Stats from {host}:{port}')
    console.print('\n\n')
    console.rule(aprsd_title)

    # Show the connection to APRS
    # It can be a connection to an APRS-IS server or a local TNC via KISS or KISSTCP
    if 'aprs-is' in stats:
        title = f'APRS-IS Connection {stats["APRSClientStats"]["server_string"]}'
        table = Table(title=title)
        table.add_column('Key')
        table.add_column('Value')
        for key, value in stats['APRSClientStats'].items():
            table.add_row(key, value)
        console.print(table)

    threads_table = Table(title='Threads')
    threads_table.add_column('Name')
    threads_table.add_column('Alive?')
    for name, alive in stats['APRSDThreadList'].items():
        threads_table.add_row(name, str(alive))

    console.print(threads_table)

    packet_totals = Table(title='Packet Totals')
    packet_totals.add_column('Key')
    packet_totals.add_column('Value')
    packet_totals.add_row('Total Received', str(stats['PacketList']['rx']))
    packet_totals.add_row('Total Sent', str(stats['PacketList']['tx']))
    console.print(packet_totals)

    # Show each of the packet types
    packets_table = Table(title='Packets By Type')
    packets_table.add_column('Packet Type')
    packets_table.add_column('TX')
    packets_table.add_column('RX')
    for key, value in stats['PacketList']['packets'].items():
        packets_table.add_row(key, str(value['tx']), str(value['rx']))

    console.print(packets_table)

    if 'plugins' in stats:
        count = len(stats['PluginManager'])
        plugins_table = Table(title=f'Plugins ({count})')
        plugins_table.add_column('Plugin')
        plugins_table.add_column('Enabled')
        plugins_table.add_column('Version')
        plugins_table.add_column('TX')
        plugins_table.add_column('RX')
        plugins = stats['PluginManager']
        for key, _ in plugins.items():
            plugins_table.add_row(
                key,
                str(plugins[key]['enabled']),
                plugins[key]['version'],
                str(plugins[key]['tx']),
                str(plugins[key]['rx']),
            )

        console.print(plugins_table)

    if seen_list := stats.get('SeenList'):
        count = len(seen_list)
        seen_table = Table(title=f'Seen List ({count})')
        seen_table.add_column('Callsign')
        seen_table.add_column('Message Count')
        seen_table.add_column('Last Heard')
        for key, value in seen_list.items():
            seen_table.add_row(key, str(value['count']), value['last'])

        console.print(seen_table)

    if watch_list := stats.get('WatchList'):
        count = len(watch_list)
        watch_table = Table(title=f'Watch List ({count})')
        watch_table.add_column('Callsign')
        watch_table.add_column('Last Heard')
        for key, value in watch_list.items():
            watch_table.add_row(key, value['last'])

        console.print(watch_table)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    '--raw',
    is_flag=True,
    default=False,
    help='Dump raw stats instead of formatted output.',
)
@click.option(
    '--show-section',
    default=['All'],
    help='Show specific sections of the stats. '
    ' Choices: All, APRSDStats, APRSDThreadList, APRSClientStats,'
    ' PacketList, SeenList, WatchList',
    multiple=True,
    type=click.Choice(
        [
            'All',
            'APRSDStats',
            'APRSDThreadList',
            'APRSClientStats',
            'PacketList',
            'SeenList',
            'WatchList',
        ],
        case_sensitive=False,
    ),
)
@click.pass_context
@cli_helper.process_standard_options
def dump_stats(ctx, raw, show_section):
    """Dump the current stats from the running APRSD instance."""
    console = Console()
    console.print(f'APRSD Dump-Stats started version: {aprsd.__version__}')

    with console.status('Dumping stats'):
        ss = StatsStore()
        ss.load()
        stats = ss.data
        if raw:
            if 'All' in show_section:
                console.print(stats)
                return
            else:
                for section in show_section:
                    console.print(f'Dumping {section} section:')
                    console.print(stats[section])
                return

        t = Table(title='APRSD Stats')
        t.add_column('Key')
        t.add_column('Value')
        for key, value in stats['APRSDStats'].items():
            t.add_row(key, str(value))

        if 'All' in show_section or 'APRSDStats' in show_section:
            console.print(t)

        # Show the thread list
        t = Table(title='Thread List')
        t.add_column('Name')
        t.add_column('Class')
        t.add_column('Alive?')
        t.add_column('Loop Count')
        t.add_column('Age')
        for name, value in stats['APRSDThreadList'].items():
            t.add_row(
                name,
                value['class'],
                str(value['alive']),
                str(value['loop_count']),
                str(value['age']),
            )

        if 'All' in show_section or 'APRSDThreadList' in show_section:
            console.print(t)

        # Show the plugins
        t = Table(title='Plugin List')
        t.add_column('Name')
        t.add_column('Enabled')
        t.add_column('Version')
        t.add_column('TX')
        t.add_column('RX')
        for name, value in stats['PluginManager'].items():
            t.add_row(
                name,
                str(value['enabled']),
                value['version'],
                str(value['tx']),
                str(value['rx']),
            )

        if 'All' in show_section or 'PluginManager' in show_section:
            console.print(t)

        # Now show the client stats
        t = Table(title='Client Stats')
        t.add_column('Key')
        t.add_column('Value')
        for key, value in stats['APRSClientStats'].items():
            t.add_row(key, str(value))

        if 'All' in show_section or 'APRSClientStats' in show_section:
            console.print(t)

        # now show the packet list
        packet_list = stats.get('PacketList')
        t = Table(title='Packet List')
        t.add_column('Key')
        t.add_column('Value')
        t.add_row('Total Received', str(packet_list['rx']))
        t.add_row('Total Sent', str(packet_list['tx']))

        if 'All' in show_section or 'PacketList' in show_section:
            console.print(t)

        # now show the seen list
        seen_list = stats.get('SeenList')
        sorted_seen_list = sorted(
            seen_list.items(),
        )
        t = Table(title='Seen List')
        t.add_column('Callsign')
        t.add_column('Message Count')
        t.add_column('Last Heard')
        for key, value in sorted_seen_list:
            t.add_row(
                key,
                str(value['count']),
                str(value['last']),
            )

        if 'All' in show_section or 'SeenList' in show_section:
            console.print(t)

        # now show the watch list
        watch_list = stats.get('WatchList')
        sorted_watch_list = sorted(
            watch_list.items(),
        )
        t = Table(title='Watch List')
        t.add_column('Callsign')
        t.add_column('Last Heard')
        for key, value in sorted_watch_list:
            t.add_row(
                key,
                str(value['last']),
            )

        if 'All' in show_section or 'WatchList' in show_section:
            console.print(t)



from aprsd.packets import collector
from aprsd.packets.core import (  # noqa: F401
    AckPacket,
    BeaconPacket,
    BulletinPacket,
    GPSPacket,
    MessagePacket,
    MicEPacket,
    ObjectPacket,
    Packet,
    RejectPacket,
    StatusPacket,
    ThirdPartyPacket,
    UnknownPacket,
    WeatherPacket,
    factory,
)
from aprsd.packets.filter import PacketFilter
from aprsd.packets.filters.dupe_filter import DupePacketFilter
from aprsd.packets.packet_list import PacketList  # noqa: F401
from aprsd.packets.seen_list import SeenList  # noqa: F401
from aprsd.packets.tracker import PacketTrack  # noqa: F401
from aprsd.packets.watch_list import WatchList  # noqa: F401

# Register all the packet tracking objects.
collector.PacketCollector().register(PacketList)
collector.PacketCollector().register(SeenList)
collector.PacketCollector().register(PacketTrack)
collector.PacketCollector().register(WatchList)

# Register all the packet filters for normal processing
# For specific commands you can deregister these if you don't want them.
PacketFilter().register(DupePacketFilter)


NULL_MESSAGE = -1



import json
import logging
import re

import requests
from oslo_config import cfg

from aprsd import plugin, plugin_utils
from aprsd.utils import trace

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class USWeatherPlugin(plugin.APRSDRegexCommandPluginBase, plugin.APRSFIKEYMixin):
    """USWeather Command

    Returns a weather report for the calling weather station
    inside the United States only.  This uses the
    forecast.weather.gov API to fetch the weather.

    This service does not require an apiKey.

    How to Call: Send a message to aprsd
    "weather" - returns weather near the calling callsign
    """

    # command_regex = r"^([w][x]|[w][x]\s|weather)"
    command_regex = r'^[wW]'

    command_name = 'USWeather'
    short_description = 'Provide USA only weather of GPS Beacon location'

    def setup(self):
        self.ensure_aprs_fi_key()

    @trace.trace
    def process(self, packet):
        LOG.info('Weather Plugin')
        fromcall = packet.from_call
        message = packet.get('message_text', None)
        # message = packet.get("message_text", None)
        # ack = packet.get("msgNo", "0")
        a = re.search(r'^.*\s+(.*)', message)
        if a is not None:
            searchcall = a.group(1)
            searchcall = searchcall.upper()
        else:
            searchcall = fromcall
        api_key = CONF.aprs_fi.apiKey
        try:
            aprs_data = plugin_utils.get_aprs_fi(api_key, searchcall)
        except Exception as ex:
            LOG.error(f'Failed to fetch aprs.fi data {ex}')
            return 'Failed to fetch aprs.fi location'

        LOG.debug(f'LocationPlugin: aprs_data = {aprs_data}')
        if not len(aprs_data['entries']):
            LOG.error("Didn't get any entries from aprs.fi")
            return 'Failed to fetch aprs.fi location'

        lat = aprs_data['entries'][0]['lat']
        lon = aprs_data['entries'][0]['lng']

        try:
            wx_data = plugin_utils.get_weather_gov_for_gps(lat, lon)
        except Exception as ex:
            LOG.error(f"Couldn't fetch forecast.weather.gov '{ex}'")
            return 'Unable to get weather'

        LOG.info(f'WX data {wx_data}')

        reply = (
            '%sF(%sF/%sF) %s. %s, %s.'
            % (
                wx_data['currentobservation']['Temp'],
                wx_data['data']['temperature'][0],
                wx_data['data']['temperature'][1],
                wx_data['data']['weather'][0],
                wx_data['time']['startPeriodName'][1],
                wx_data['data']['weather'][1],
            )
        ).rstrip()
        LOG.debug(f"reply: '{reply}' ")
        return reply


class USMetarPlugin(plugin.APRSDRegexCommandPluginBase, plugin.APRSFIKEYMixin):
    """METAR Command

    This provides a METAR weather report from a station near the caller
    or callsign using the forecast.weather.gov api.  This only works
    for stations inside the United States.

    This service does not require an apiKey.

    How to Call: Send a message to aprsd
    "metar" - returns metar report near the calling callsign
    "metar CALLSIGN" - returns metar report near CALLSIGN

    """

    command_regex = r'^([m]|[M]|[m]\s|metar)'
    command_name = 'USMetar'
    short_description = 'USA only METAR of GPS Beacon location'

    def setup(self):
        self.ensure_aprs_fi_key()

    @trace.trace
    def process(self, packet):
        fromcall = packet.get('from')
        message = packet.get('message_text', None)
        # ack = packet.get("msgNo", "0")
        LOG.info(f"WX Plugin '{message}'")
        a = re.search(r'^.*\s+(.*)', message)
        if a is not None:
            searchcall = a.group(1)
            station = searchcall.upper()
            try:
                resp = plugin_utils.get_weather_gov_metar(station)
            except Exception as e:
                LOG.debug(f'Weather failed with:  {str(e)}')
                reply = 'Unable to find station METAR'
            else:
                station_data = json.loads(resp.text)
                reply = station_data['properties']['rawMessage']

            return reply
        else:
            # if no second argument, search for calling station
            fromcall = fromcall

            api_key = CONF.aprs_fi.apiKey

            try:
                aprs_data = plugin_utils.get_aprs_fi(api_key, fromcall)
            except Exception as ex:
                LOG.error(f'Failed to fetch aprs.fi data {ex}')
                return 'Failed to fetch aprs.fi location'

            # LOG.debug("LocationPlugin: aprs_data = {}".format(aprs_data))
            if not len(aprs_data['entries']):
                LOG.error('Found no entries from aprs.fi!')
                return 'Failed to fetch aprs.fi location'

            lat = aprs_data['entries'][0]['lat']
            lon = aprs_data['entries'][0]['lng']

            try:
                wx_data = plugin_utils.get_weather_gov_for_gps(lat, lon)
            except Exception as ex:
                LOG.error(f"Couldn't fetch forecast.weather.gov '{ex}'")
                return 'Unable to metar find station.'

            if wx_data['location']['metar']:
                station = wx_data['location']['metar']
                try:
                    resp = plugin_utils.get_weather_gov_metar(station)
                except Exception as e:
                    LOG.debug(f'Weather failed with:  {str(e)}')
                    reply = 'Failed to get Metar'
                else:
                    station_data = json.loads(resp.text)
                    reply = station_data['properties']['rawMessage']
            else:
                # Couldn't find a station
                reply = 'No Metar station found'

        return reply


class OWMWeatherPlugin(plugin.APRSDRegexCommandPluginBase):
    """OpenWeatherMap Weather Command

    This provides weather near the caller or callsign.

    How to Call: Send a message to aprsd
    "weather" - returns the weather near the calling callsign
    "weather CALLSIGN" - returns the weather near CALLSIGN

    This plugin uses the openweathermap API to fetch
    location and weather information.

    To use this plugin you need to get an openweathermap
    account and apikey.

    https://home.openweathermap.org/api_keys

    """

    # command_regex = r"^([w][x]|[w][x]\s|weather)"
    command_regex = r'^[wW]'

    command_name = 'OpenWeatherMap'
    short_description = 'OpenWeatherMap weather of GPS Beacon location'

    def setup(self):
        if not CONF.owm_weather_plugin.apiKey:
            LOG.error('Config.owm_weather_plugin.apiKey is not set.  Disabling')
            self.enabled = False
        else:
            self.enabled = True

    def help(self):
        _help = [
            'openweathermap: Send {} to get weather from your location'.format(
                self.command_regex
            ),
            'openweathermap: Send {} <callsign> to get weather from <callsign>'.format(
                self.command_regex
            ),
        ]
        return _help

    @trace.trace
    def process(self, packet):
        fromcall = packet.get('from_call')
        message = packet.get('message_text', None)
        # ack = packet.get("msgNo", "0")
        LOG.info(f"OWMWeather Plugin '{message}'")
        a = re.search(r'^.*\s+(.*)', message)
        if a is not None:
            searchcall = a.group(1)
            searchcall = searchcall.upper()
        else:
            searchcall = fromcall

        api_key = CONF.aprs_fi.apiKey

        try:
            aprs_data = plugin_utils.get_aprs_fi(api_key, searchcall)
        except Exception as ex:
            LOG.error(f'Failed to fetch aprs.fi data {ex}')
            return 'Failed to fetch location'

        # LOG.debug("LocationPlugin: aprs_data = {}".format(aprs_data))
        if not len(aprs_data['entries']):
            LOG.error('Found no entries from aprs.fi!')
            return 'Failed to fetch location'

        lat = aprs_data['entries'][0]['lat']
        lon = aprs_data['entries'][0]['lng']

        units = CONF.units
        api_key = CONF.owm_weather_plugin.apiKey
        try:
            wx_data = plugin_utils.fetch_openweathermap(
                api_key,
                lat,
                lon,
                units=units,
                exclude='minutely,hourly',
            )
        except Exception as ex:
            LOG.error(f"Couldn't fetch openweathermap api '{ex}'")
            # default to UTC
            return 'Unable to get weather'

        if units == 'metric':
            degree = 'C'
        else:
            degree = 'F'

        if 'wind_gust' in wx_data['current']:
            wind = '{:.0f}@{}G{:.0f}'.format(
                wx_data['current']['wind_speed'],
                wx_data['current']['wind_deg'],
                wx_data['current']['wind_gust'],
            )
        else:
            wind = '{:.0f}@{}'.format(
                wx_data['current']['wind_speed'],
                wx_data['current']['wind_deg'],
            )

        # LOG.debug(wx_data["current"])
        # LOG.debug(wx_data["daily"])
        reply = '{} {:.1f}{}/{:.1f}{} Wind {} {}%'.format(
            wx_data['current']['weather'][0]['description'],
            wx_data['current']['temp'],
            degree,
            wx_data['current']['dew_point'],
            degree,
            wind,
            wx_data['current']['humidity'],
        )

        return reply


class AVWXWeatherPlugin(plugin.APRSDRegexCommandPluginBase):
    """AVWXWeatherMap Weather Command

    Fetches a METAR weather report for the nearest
    weather station from the callsign
    Can be called with:
    metar - fetches metar for caller
    metar <CALLSIGN> - fetches metar for <CALLSIGN>

    This plugin requires the avwx-api service
    to provide the metar for a station near
    the callsign.

    avwx-api is an opensource project that has
    a hosted service here: https://avwx.rest/

    You can launch your own avwx-api in a container
    by cloning the githug repo here: https://github.com/avwx-rest/AVWX-API

    Then build the docker container with:
    docker build -f Dockerfile -t avwx-api:master .
    """

    command_regex = r'^([m]|[m]|[m]\s|metar)'
    command_name = 'AVWXWeather'
    short_description = 'AVWX weather of GPS Beacon location'

    def setup(self):
        if not CONF.avwx_plugin.base_url:
            LOG.error('Config avwx_plugin.base_url not specified.  Disabling')
            return False
        elif not CONF.avwx_plugin.apiKey:
            LOG.error('Config avwx_plugin.apiKey not specified. Disabling')
            return False

        self.enabled = True

    def help(self):
        _help = [
            'avwxweather: Send {} to get weather from your location'.format(
                self.command_regex
            ),
            'avwxweather: Send {} <callsign> to get weather from <callsign>'.format(
                self.command_regex
            ),
        ]
        return _help

    @trace.trace
    def process(self, packet):
        fromcall = packet.get('from')
        message = packet.get('message_text', None)
        # ack = packet.get("msgNo", "0")
        LOG.info(f"AVWXWeather Plugin '{message}'")
        a = re.search(r'^.*\s+(.*)', message)
        if a is not None:
            searchcall = a.group(1)
            searchcall = searchcall.upper()
        else:
            searchcall = fromcall

        api_key = CONF.aprs_fi.apiKey
        try:
            aprs_data = plugin_utils.get_aprs_fi(api_key, searchcall)
        except Exception as ex:
            LOG.error(f'Failed to fetch aprs.fi data {ex}')
            return 'Failed to fetch location'

        # LOG.debug("LocationPlugin: aprs_data = {}".format(aprs_data))
        if not len(aprs_data['entries']):
            LOG.error('Found no entries from aprs.fi!')
            return 'Failed to fetch location'

        lat = aprs_data['entries'][0]['lat']
        lon = aprs_data['entries'][0]['lng']

        api_key = CONF.avwx_plugin.apiKey
        base_url = CONF.avwx_plugin.base_url
        token = f'TOKEN {api_key}'
        headers = {'Authorization': token}
        try:
            coord = f'{lat},{lon}'
            url = (
                '{}/api/station/near/{}?'
                'n=1&airport=false&reporting=true&format=json'.format(base_url, coord)
            )

            LOG.debug(f"Get stations near me '{url}'")
            response = requests.get(url, headers=headers)
        except Exception as ex:
            LOG.error(ex)
            raise Exception(f"Failed to get the weather '{ex}'") from ex
        else:
            wx_data = json.loads(response.text)

        # LOG.debug(wx_data)
        station = wx_data[0]['station']['icao']

        try:
            url = (
                '{}/api/metar/{}?options=info,translate,summary'
                '&airport=true&reporting=true&format=json&onfail=cache'.format(
                    base_url,
                    station,
                )
            )

            LOG.debug(f"Get METAR '{url}'")
            response = requests.get(url, headers=headers)
        except Exception as ex:
            LOG.error(ex)
            raise Exception(f'Failed to get metar {ex}') from ex
        else:
            metar_data = json.loads(response.text)

        # LOG.debug(metar_data)
        return metar_data['raw']



import logging
import threading
import time

import wrapt
from oslo_config import cfg
from rush import quota, throttle
from rush.contrib import decorator
from rush.limiters import periodic
from rush.stores import dictionary

from aprsd import conf  # noqa
from aprsd import threads as aprsd_threads
from aprsd.client.client import APRSDClient
from aprsd.packets import collector, core, tracker
from aprsd.packets import log as packet_log

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')

msg_t = throttle.Throttle(
    limiter=periodic.PeriodicLimiter(
        store=dictionary.DictionaryStore(),
    ),
    rate=quota.Quota.per_second(
        count=CONF.msg_rate_limit_period,
    ),
)
ack_t = throttle.Throttle(
    limiter=periodic.PeriodicLimiter(
        store=dictionary.DictionaryStore(),
    ),
    rate=quota.Quota.per_second(
        count=CONF.ack_rate_limit_period,
    ),
)

msg_throttle_decorator = decorator.ThrottleDecorator(throttle=msg_t)
ack_throttle_decorator = decorator.ThrottleDecorator(throttle=ack_t)
s_lock = threading.Lock()


@wrapt.synchronized(s_lock)
@msg_throttle_decorator.sleep_and_retry
def send(packet: core.Packet, direct=False, aprs_client=None):
    """Send a packet either in a thread or directly to the client."""
    # prepare the packet for sending.
    # This constructs the packet.raw
    packet.prepare(create_msg_number=True)
    # Have to call the collector to track the packet
    # After prepare, as prepare assigns the msgNo
    collector.PacketCollector().tx(packet)
    if isinstance(packet, core.AckPacket):
        if CONF.enable_sending_ack_packets:
            _send_ack(packet, direct=direct, aprs_client=aprs_client)
        else:
            LOG.info('Sending ack packets is disabled. Not sending AckPacket.')
    else:
        _send_packet(packet, direct=direct, aprs_client=aprs_client)


@msg_throttle_decorator.sleep_and_retry
def _send_packet(packet: core.Packet, direct=False, aprs_client=None):
    if not direct:
        thread = SendPacketThread(packet=packet)
        thread.start()
    else:
        _send_direct(packet, aprs_client=aprs_client)


@ack_throttle_decorator.sleep_and_retry
def _send_ack(packet: core.AckPacket, direct=False, aprs_client=None):
    if not direct:
        thread = SendAckThread(packet=packet)
        thread.start()
    else:
        _send_direct(packet, aprs_client=aprs_client)


def _send_direct(packet, aprs_client=None):
    if aprs_client:
        cl = aprs_client
    else:
        cl = APRSDClient()

    packet_log.log(packet, tx=True)
    try:
        cl.send(packet)
    except Exception as e:
        LOG.error(f'Failed to send packet: {packet}')
        LOG.error(e)
        return False
    else:
        return True


class SendPacketThread(aprsd_threads.APRSDThread):
    loop_count: int = 1

    def __init__(self, packet):
        self.packet = packet
        super().__init__(f'TX-{packet.to_call}-{self.packet.msgNo}')

    def loop(self):
        """Loop until a message is acked or it gets delayed.

        We only sleep for 5 seconds between each loop run, so
        that CTRL-C can exit the app in a short period.  Each sleep
        means the app quitting is blocked until sleep is done.
        So we keep track of the last send attempt and only send if the
        last send attempt is old enough.

        """
        pkt_tracker = tracker.PacketTrack()
        # lets see if the message is still in the tracking queue
        packet = pkt_tracker.get(self.packet.msgNo)
        if not packet:
            # The message has been removed from the tracking queue
            # So it got acked and we are done.
            LOG.info(
                f'{self.packet.__class__.__name__}'
                f'({self.packet.msgNo}) '
                'Message Send Complete via Ack.',
            )
            return False
        else:
            send_now = False
            if packet.send_count >= packet.retry_count:
                # we reached the send limit, don't send again
                # TODO(hemna) - Need to put this in a delayed queue?
                LOG.info(
                    f'{packet.__class__.__name__} '
                    f'({packet.msgNo}) '
                    'Message Send Complete. Max attempts reached'
                    f' {packet.retry_count}',
                )
                pkt_tracker.remove(packet.msgNo)
                return False

            # Message is still outstanding and needs to be acked.
            if packet.last_send_time:
                # Message has a last send time tracking
                now = int(round(time.time()))
                sleeptime = (packet.send_count + 1) * 31
                delta = now - packet.last_send_time
                if delta > sleeptime:
                    # It's time to try to send it again
                    send_now = True
            else:
                send_now = True

            if send_now:
                # no attempt time, so lets send it, and start
                # tracking the time.
                packet.last_send_time = int(round(time.time()))
                sent = False
                try:
                    sent = _send_direct(packet)
                except Exception as ex:
                    LOG.error(f'Failed to send packet: {packet}')
                    LOG.error(ex)
                else:
                    # If an exception happens while sending
                    # we don't want this attempt to count
                    # against the packet
                    if sent:
                        packet.send_count += 1

            time.sleep(1)
            # Make sure we get called again.
            self.loop_count += 1
            return True


class SendAckThread(aprsd_threads.APRSDThread):
    loop_count: int = 1
    max_retries = 3

    def __init__(self, packet):
        self.packet = packet
        super().__init__(f'TXAck-{packet.to_call}-{self.packet.msgNo}')
        self.max_retries = CONF.default_ack_send_count

    def loop(self):
        """Separate thread to send acks with retries."""
        send_now = False
        if self.packet.send_count == self.max_retries:
            # we reached the send limit, don't send again
            # TODO(hemna) - Need to put this in a delayed queue?
            LOG.debug(
                f'{self.packet.__class__.__name__}'
                f'({self.packet.msgNo}) '
                'Send Complete. Max attempts reached'
                f' {self.max_retries}',
            )
            return False

        if self.packet.last_send_time:
            # Message has a last send time tracking
            now = int(round(time.time()))

            # aprs duplicate detection is 30 secs?
            # (21 only sends first, 28 skips middle)
            sleep_time = 31
            delta = now - self.packet.last_send_time
            if delta > sleep_time:
                # It's time to try to send it again
                send_now = True
            elif self.loop_count % 10 == 0:
                # LOG.debug(f'Still wating. {delta}')
                pass
        else:
            send_now = True

        if send_now:
            sent = False
            try:
                sent = _send_direct(self.packet)
            except Exception:
                LOG.error(f'Failed to send packet: {self.packet}')
            else:
                # If an exception happens while sending
                # we don't want this attempt to count
                # against the packet
                if sent:
                    self.packet.send_count += 1

            self.packet.last_send_time = int(round(time.time()))

        time.sleep(1)
        self.loop_count += 1
        return True


class BeaconSendThread(aprsd_threads.APRSDThread):
    """Thread that sends a GPS beacon packet periodically.

    Settings are in the [DEFAULT] section of the config file.
    """

    _loop_cnt: int = 1

    def __init__(self):
        super().__init__('BeaconSendThread')
        self._loop_cnt = 1
        # Make sure Latitude and Longitude are set.
        if not CONF.latitude or not CONF.longitude:
            LOG.error(
                'Latitude and Longitude are not set in the config file.'
                'Beacon will not be sent and thread is STOPPED.',
            )
            self.stop()
        LOG.info(
            'Beacon thread is running and will send '
            f'beacons every {CONF.beacon_interval} seconds.',
        )

    def loop(self):
        # Only dump out the stats every N seconds
        if self._loop_cnt % CONF.beacon_interval == 0:
            pkt = core.BeaconPacket(
                from_call=CONF.callsign,
                to_call='APRS',
                latitude=float(CONF.latitude),
                longitude=float(CONF.longitude),
                comment='APRSD GPS Beacon',
                symbol=CONF.beacon_symbol,
            )
            try:
                # Only send it once
                pkt.retry_count = 1
                send(pkt, direct=True)
            except Exception as e:
                LOG.error(f'Failed to send beacon: {e}')
                APRSDClient().reset()
                time.sleep(5)

        self._loop_cnt += 1
        time.sleep(1)
        return True



# All client drivers must be registered here
from aprsd.client.drivers.aprsis import APRSISDriver
from aprsd.client.drivers.fake import APRSDFakeDriver
from aprsd.client.drivers.registry import DriverRegistry
from aprsd.client.drivers.tcpkiss import TCPKISSDriver

driver_registry = DriverRegistry()
driver_registry.register(APRSDFakeDriver)
driver_registry.register(APRSISDriver)
driver_registry.register(TCPKISSDriver)



import logging
from typing import Union

from oslo_config import cfg

from aprsd import packets
from aprsd.packets import core

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class DupePacketFilter:
    """This is a packet filter to detect duplicate packets.

    This Uses the PacketList object to see if a packet exists
    already.  If it does exist in the PacketList, then we need to
    check the flag on the packet to see if it's been processed before.
    If the packet has been processed already within the allowed
    timeframe, then it's a dupe.
    """

    def filter(self, packet: type[core.Packet]) -> Union[type[core.Packet], None]:
        # LOG.debug(f"{self.__class__.__name__}.filter called for packet {packet}")
        """Filter a packet out if it's already been seen and processed."""
        if isinstance(packet, core.AckPacket):
            # We don't need to drop AckPackets, those should be
            # processed.
            # Send the AckPacket to the queue for processing elsewhere.
            return packet
        else:
            # Make sure we aren't re-processing the same packet
            # For RF based APRS Clients we can get duplicate packets
            # So we need to track them and not process the dupes.
            pkt_list = packets.PacketList()
            found = False
            try:
                # Find the packet in the list of already seen packets
                # Based on the packet.key
                found = pkt_list.find(packet)
                if not packet.msgNo:
                    # If the packet doesn't have a message id
                    # then there is no reliable way to detect
                    # if it's a dupe, so we just pass it on.
                    # it shouldn't get acked either.
                    found = False
            except KeyError:
                found = False

            if not found:
                # We haven't seen this packet before, so we process it.
                return packet

            if not packet.processed:
                # We haven't processed this packet through the plugins.
                return packet
            elif packet.timestamp - found.timestamp < CONF.packet_dupe_timeout:
                # If the packet came in within N seconds of the
                # Last time seeing the packet, then we drop it as a dupe.
                LOG.warning(
                    f'Packet {packet.from_call}:{packet.msgNo} already tracked, dropping.'
                )
            else:
                LOG.warning(
                    f'Packet {packet.from_call}:{packet.msgNo} already tracked '
                    f'but older than {CONF.packet_dupe_timeout} seconds. processing.',
                )
                return packet



import abc
import logging
import queue
import time

import aprslib
from oslo_config import cfg

from aprsd import packets, plugin
from aprsd.client.client import APRSDClient
from aprsd.packets import collector, filter
from aprsd.packets import log as packet_log
from aprsd.threads import APRSDThread, tx

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class APRSDRXThread(APRSDThread):
    """Main Class to connect to an APRS Client and recieve packets.

    A packet is received in the main loop and then sent to the
    process_packet method, which sends the packet through the collector
    to track the packet for stats, and then put into the packet queue
    for processing in a separate thread.
    """

    _client = None

    # This is the queue that packets are sent to for processing.
    # We process packets in a separate thread to help prevent
    # getting blocked by the APRS server trying to send us packets.
    packet_queue = None

    pkt_count = 0

    def __init__(self, packet_queue):
        super().__init__('RX_PKT')
        self.packet_queue = packet_queue

    def stop(self):
        self.thread_stop = True
        if self._client:
            self._client.close()

    def loop(self):
        if not self._client:
            self._client = APRSDClient()
            time.sleep(1)
            return True

        if not self._client.is_alive:
            self._client = APRSDClient()
            time.sleep(1)
            return True

        # setup the consumer of messages and block until a messages
        try:
            # This will register a packet consumer with aprslib
            # When new packets come in the consumer will process
            # the packet

            # Do a partial here because the consumer signature doesn't allow
            # For kwargs to be passed in to the consumer func we declare
            # and the aprslib developer didn't want to allow a PR to add
            # kwargs.  :(
            # https://github.com/rossengeorgiev/aprs-python/pull/56
            self._client.consumer(
                self.process_packet,
                raw=False,
            )
        except (
            aprslib.exceptions.ConnectionDrop,
            aprslib.exceptions.ConnectionError,
        ):
            LOG.error('Connection dropped, reconnecting')
            # Force the deletion of the client object connected to aprs
            # This will cause a reconnect, next time client.get_client()
            # is called
            self._client.reset()
            time.sleep(5)
        except Exception as ex:
            LOG.exception(ex)
            LOG.error('Resetting connection and trying again.')
            self._client.reset()
            time.sleep(5)
        return True

    def process_packet(self, *args, **kwargs):
        packet = self._client.decode_packet(*args, **kwargs)
        if not packet:
            LOG.error(
                'No packet received from decode_packet.  Most likely a failure to parse'
            )
            return
        self.pkt_count += 1
        packet_log.log(packet, packet_count=self.pkt_count)
        pkt_list = packets.PacketList()

        if isinstance(packet, packets.AckPacket):
            # We don't need to drop AckPackets, those should be
            # processed.
            self.packet_queue.put(packet)
        else:
            # Make sure we aren't re-processing the same packet
            # For RF based APRS Clients we can get duplicate packets
            # So we need to track them and not process the dupes.
            found = False
            try:
                # Find the packet in the list of already seen packets
                # Based on the packet.key
                found = pkt_list.find(packet)
                if not packet.msgNo:
                    # If the packet doesn't have a message id
                    # then there is no reliable way to detect
                    # if it's a dupe, so we just pass it on.
                    # it shouldn't get acked either.
                    found = False
            except KeyError:
                found = False

            if not found:
                # We haven't seen this packet before, so we process it.
                collector.PacketCollector().rx(packet)
                self.packet_queue.put(packet)
            elif packet.timestamp - found.timestamp < CONF.packet_dupe_timeout:
                # If the packet came in within N seconds of the
                # Last time seeing the packet, then we drop it as a dupe.
                LOG.warning(
                    f'Packet {packet.from_call}:{packet.msgNo} already tracked, dropping.'
                )
            else:
                LOG.warning(
                    f'Packet {packet.from_call}:{packet.msgNo} already tracked '
                    f'but older than {CONF.packet_dupe_timeout} seconds. processing.',
                )
                collector.PacketCollector().rx(packet)
                self.packet_queue.put(packet)


class APRSDFilterThread(APRSDThread):
    def __init__(self, thread_name, packet_queue):
        super().__init__(thread_name)
        self.packet_queue = packet_queue

    def filter_packet(self, packet):
        # Do any packet filtering prior to processing
        if not filter.PacketFilter().filter(packet):
            return None
        return packet

    def print_packet(self, packet):
        """Allow a child of this class to override this.

        This is helpful if for whatever reason the child class
        doesn't want to log packets.

        """
        packet_log.log(packet)

    def loop(self):
        try:
            packet = self.packet_queue.get(timeout=1)
            self.print_packet(packet)
            if packet:
                if self.filter_packet(packet):
                    self.process_packet(packet)
        except queue.Empty:
            pass
        return True


class APRSDProcessPacketThread(APRSDFilterThread):
    """Base class for processing received packets after they have been filtered.

    Packets are received from the client, then filtered for dupes,
    then sent to the packet queue.  This thread pulls packets from
    the packet queue for processing.

    This is the base class for processing packets coming from
    the consumer.  This base class handles sending ack packets and
    will ack a message before sending the packet to the subclass
    for processing."""

    def __init__(self, packet_queue):
        super().__init__('ProcessPKT', packet_queue=packet_queue)
        if not CONF.enable_sending_ack_packets:
            LOG.warning(
                'Sending ack packets is disabled, messages will not be acknowledged.',
            )

    def process_ack_packet(self, packet):
        """We got an ack for a message, no need to resend it."""
        ack_num = packet.msgNo
        LOG.debug(f'Got ack for message {ack_num}')
        collector.PacketCollector().rx(packet)

    def process_piggyback_ack(self, packet):
        """We got an ack embedded in a packet."""
        ack_num = packet.ackMsgNo
        LOG.debug(f'Got PiggyBackAck for message {ack_num}')
        collector.PacketCollector().rx(packet)

    def process_reject_packet(self, packet):
        """We got a reject message for a packet.  Stop sending the message."""
        ack_num = packet.msgNo
        LOG.debug(f'Got REJECT for message {ack_num}')
        collector.PacketCollector().rx(packet)

    def process_packet(self, packet):
        """Process a packet received from aprs-is server."""
        LOG.debug(f'ProcessPKT-LOOP {self.loop_count}')

        # set this now as we are going to process it.
        # This is used during dupe checking, so set it early
        packet.processed = True

        our_call = CONF.callsign.lower()

        from_call = packet.from_call
        if hasattr(packet, 'addresse') and packet.addresse:
            to_call = packet.addresse
        else:
            to_call = packet.to_call
        msg_id = packet.msgNo

        # We don't put ack packets destined for us through the
        # plugins.
        if (
            isinstance(packet, packets.AckPacket)
            and packet.addresse.lower() == our_call
        ):
            self.process_ack_packet(packet)
        elif (
            isinstance(packet, packets.RejectPacket)
            and packet.addresse.lower() == our_call
        ):
            self.process_reject_packet(packet)
        else:
            if hasattr(packet, 'ackMsgNo') and packet.ackMsgNo:
                # we got an ack embedded in this packet
                # we need to handle the ack
                self.process_piggyback_ack(packet)
            # Only ack messages that were sent directly to us
            if isinstance(packet, packets.MessagePacket):
                if to_call and to_call.lower() == our_call:
                    # It's a MessagePacket and it's for us!
                    # let any threads do their thing, then ack
                    # send an ack last
                    if msg_id:
                        tx.send(
                            packets.AckPacket(
                                from_call=CONF.callsign,
                                to_call=from_call,
                                msgNo=msg_id,
                            ),
                        )

                    self.process_our_message_packet(packet)
                else:
                    # Packet wasn't meant for us!
                    self.process_other_packet(packet, for_us=False)
            else:
                self.process_other_packet(
                    packet,
                    for_us=(to_call.lower() == our_call),
                )
        LOG.debug(f"Packet processing complete for pkt '{packet.key}'")
        return False

    @abc.abstractmethod
    def process_our_message_packet(self, packet):
        """Process a MessagePacket destined for us!"""

    def process_other_packet(self, packet, for_us=False):
        """Process an APRS Packet that isn't a message or ack"""
        if not for_us:
            LOG.info(f"Got a packet meant for someone else '{packet.to_call}'")
        else:
            LOG.info('Got a non AckPacket/MessagePacket')


class APRSDPluginProcessPacketThread(APRSDProcessPacketThread):
    """Process the packet through the plugin manager.

    This is the main aprsd server plugin processing thread."""

    def process_other_packet(self, packet, for_us=False):
        pm = plugin.PluginManager()
        try:
            results = pm.run_watchlist(packet)
            for reply in results:
                if isinstance(reply, list):
                    for subreply in reply:
                        LOG.debug(f"Sending '{subreply}'")
                        if isinstance(subreply, packets.Packet):
                            tx.send(subreply)
                        else:
                            wl = CONF.watch_list
                            to_call = wl['alert_callsign']
                            tx.send(
                                packets.MessagePacket(
                                    from_call=CONF.callsign,
                                    to_call=to_call,
                                    message_text=subreply,
                                ),
                            )
                elif isinstance(reply, packets.Packet):
                    # We have a message based object.
                    tx.send(reply)
        except Exception as ex:
            LOG.error('Plugin failed!!!')
            LOG.exception(ex)

    def process_our_message_packet(self, packet):
        """Send the packet through the plugins."""
        from_call = packet.from_call
        if packet.addresse:
            to_call = packet.addresse
        else:
            to_call = None

        pm = plugin.PluginManager()
        try:
            results = pm.run(packet)
            replied = False
            for reply in results:
                if isinstance(reply, list):
                    # one of the plugins wants to send multiple messages
                    replied = True
                    for subreply in reply:
                        LOG.debug(f"Sending '{subreply}'")
                        if isinstance(subreply, packets.Packet):
                            tx.send(subreply)
                        else:
                            tx.send(
                                packets.MessagePacket(
                                    from_call=CONF.callsign,
                                    to_call=from_call,
                                    message_text=subreply,
                                ),
                            )
                elif isinstance(reply, packets.Packet):
                    # We have a message based object.
                    tx.send(reply)
                    replied = True
                else:
                    replied = True
                    # A plugin can return a null message flag which signals
                    # us that they processed the message correctly, but have
                    # nothing to reply with, so we avoid replying with a
                    # usage string
                    if reply is not packets.NULL_MESSAGE:
                        LOG.debug(f"Sending '{reply}'")
                        tx.send(
                            packets.MessagePacket(
                                from_call=CONF.callsign,
                                to_call=from_call,
                                message_text=reply,
                            ),
                        )

            # If the message was for us and we didn't have a
            # response, then we send a usage statement.
            if to_call == CONF.callsign and not replied:
                # Tailor the messages accordingly
                if CONF.load_help_plugin:
                    LOG.warning('Sending help!')
                    message_text = "Unknown command! Send 'help' message for help"
                else:
                    LOG.warning('Unknown command!')
                    message_text = 'Unknown command!'

                tx.send(
                    packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=from_call,
                        message_text=message_text,
                    ),
                )
        except Exception as ex:
            LOG.error('Plugin failed!!!')
            LOG.exception(ex)
            # Do we need to send a reply?
            if to_call == CONF.callsign:
                reply = 'A Plugin failed! try again?'
                tx.send(
                    packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=from_call,
                        message_text=reply,
                    ),
                )

        LOG.debug('Completed process_our_message_packet')



import logging

from oslo_config import cfg

from aprsd import packets, plugin

CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class NotifySeenPlugin(plugin.APRSDWatchListPluginBase):
    """Notification plugin to send seen message for callsign.


    This plugin will track callsigns in the watch list and report
    when a callsign has been seen when the last time they were
    seen was older than the configured age limit.
    """

    short_description = "Notify me when a CALLSIGN is recently seen on APRS-IS"

    def process(self, packet: packets.MessagePacket):
        LOG.info("NotifySeenPlugin")

        notify_callsign = CONF.watch_list.alert_callsign
        fromcall = packet.from_call

        wl = packets.WatchList()
        age = wl.age(fromcall)

        if fromcall != notify_callsign:
            if wl.is_old(fromcall):
                LOG.info(
                    "NOTIFY {} last seen {} max age={}".format(
                        fromcall,
                        age,
                        wl.max_delta(),
                    ),
                )
                packet_type = packet.__class__.__name__
                # we shouldn't notify the alert user that they are online.
                pkt = packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=notify_callsign,
                    message_text=(f"{fromcall} was just seen by type:'{packet_type}'"),
                    allow_delay=False,
                )
                pkt.allow_delay = False
                return pkt
            else:
                LOG.debug(
                    "Not old enough to notify on callsign "
                    f"'{fromcall}' : {age} < {wl.max_delta()}",
                )
                return packets.NULL_MESSAGE
        else:
            LOG.debug("fromcall and notify_callsign are the same, ignoring")
            return packets.NULL_MESSAGE



from unittest import mock

from aprsd.packets import core


class MockClientDriver:
    """Mock implementation of ClientDriver for testing."""

    def __init__(self, enabled=True, configured=True):
        self.connected = False
        self._alive = True
        self._keepalive = None
        self.filter = None
        self._enabled = enabled
        self._configured = configured
        self.path = '/dev/ttyUSB0'
        self.login_status = {
            'success': True,
            'message': None,
        }

    @staticmethod
    def is_enabled():
        """Static method to check if driver is enabled."""
        return True

    @staticmethod
    def is_configured():
        """Static method to check if driver is configured."""
        return True

    def is_alive(self):
        """Instance method to check if driver is alive."""
        return self._alive

    def stats(self, serializable=False):
        """Return mock stats."""
        stats = {'packets_received': 0, 'packets_sent': 0}
        if serializable:
            stats['path'] = self.path
        return stats

    @property
    def login_success(self):
        """Property to get login success status."""
        return self.login_status['success']

    @property
    def login_failure(self):
        """Property to get login failure message."""
        return self.login_status['message']

    def decode_packet(self, *args, **kwargs):
        """Mock packet decoding."""
        packet = mock.MagicMock(spec=core.Packet)
        packet.raw = 'test packet'
        return packet

    def close(self):
        self.connected = False

    def setup_connection(self):
        self.connected = True

    def send(self, packet):
        return True

    def set_filter(self, filter_str):
        self.filter = filter_str

    @property
    def keepalive(self):
        return self._keepalive

    def consumer(self, callback, raw=False):
        pass



from unittest import mock

import pytz
from oslo_config import cfg

from aprsd.plugins import time as time_plugin
from aprsd.utils import fuzzy

from .. import fake, test_plugin

CONF = cfg.CONF


class TestTimePlugins(test_plugin.TestPlugin):
    @mock.patch("aprsd.plugins.time.TimePlugin._get_local_tz")
    @mock.patch("aprsd.plugins.time.TimePlugin._get_utcnow")
    def test_time(self, mock_utcnow, mock_localtz):
        utcnow = pytz.datetime.datetime.utcnow()
        mock_utcnow.return_value = utcnow
        tz = pytz.timezone("US/Pacific")
        mock_localtz.return_value = tz

        gmt_t = pytz.utc.localize(utcnow)
        local_t = gmt_t.astimezone(tz)

        fake_time = mock.MagicMock()
        h = int(local_t.strftime("%H"))
        m = int(local_t.strftime("%M"))
        fake_time.tm_sec = 13
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        time = time_plugin.TimePlugin()

        packet = fake.fake_packet(
            message="location",
            msg_number=1,
        )

        actual = time.filter(packet)
        self.assertEqual(None, actual)

        cur_time = fuzzy(h, m, 1)

        packet = fake.fake_packet(
            message="time",
            msg_number=1,
        )
        local_short_str = local_t.strftime("%H:%M %Z")
        expected = "{} ({})".format(
            cur_time,
            local_short_str,
        )
        actual = time.filter(packet)
        self.assertEqual(expected, actual)



import datetime
import logging
import threading
import time
from typing import Callable

import aprslib
import wrapt
from oslo_config import cfg

from aprsd import conf  # noqa
from aprsd.packets import core
from aprsd.utils import trace

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class APRSDFakeDriver(metaclass=trace.TraceWrapperMetaclass):
    """Fake client for testing."""

    # flag to tell us to stop
    thread_stop = False

    # date for last time we heard from the server
    aprsd_keepalive = datetime.datetime.now()

    lock = threading.Lock()
    path = []

    def __init__(self):
        LOG.info('Starting APRSDFakeDriver driver.')
        self.path = ['WIDE1-1', 'WIDE2-1']

    @staticmethod
    def is_enabled():
        if CONF.fake_client.enabled:
            return True
        return False

    @staticmethod
    def is_configured():
        return APRSDFakeDriver.is_enabled

    def is_alive(self):
        """If the connection is alive or not."""
        return not self.thread_stop

    def close(self):
        self.thread_stop = True
        LOG.info('Shutdown APRSDFakeDriver driver.')

    def setup_connection(self):
        # It's fake....
        pass

    def set_filter(self, filter: str) -> None:
        pass

    def login_success(self) -> bool:
        return True

    def login_failure(self) -> str:
        return None

    @wrapt.synchronized(lock)
    def send(self, packet: core.Packet) -> bool:
        """Send an APRS Message object."""
        LOG.info(f'Sending packet: {packet}')
        payload = None
        if isinstance(packet, core.Packet):
            packet.prepare()
            payload = packet.payload.encode('US-ASCII')
        else:
            msg_payload = f'{packet.raw}{{{str(packet.msgNo)}'
            payload = (
                ':{:<9}:{}'.format(
                    packet.to_call,
                    msg_payload,
                )
            ).encode('US-ASCII')

        LOG.debug(
            f"FAKE::Send '{payload}' TO '{packet.to_call}' From "
            f'\'{packet.from_call}\' with PATH "{self.path}"',
        )
        return True

    def consumer(self, callback: Callable, raw: bool = False):
        LOG.debug('Start non blocking FAKE consumer')
        # Generate packets here?
        raw_str = 'GTOWN>APDW16,WIDE1-1,WIDE2-1:}KM6LYW-9>APZ100,TCPIP,GTOWN*::KM6LYW   :KM6LYW: 19 Miles SW'
        self.aprsd_keepalive = datetime.datetime.now()
        if raw:
            callback(raw=raw_str)
        else:
            pkt_raw = aprslib.parse(raw_str)
            pkt = core.factory(pkt_raw)
            callback(packet=pkt)

        LOG.debug(f'END blocking FAKE consumer {self}')
        time.sleep(1)

    def decode_packet(self, *args, **kwargs):
        """APRS lib already decodes this."""
        if not kwargs:
            return None

        if kwargs.get('packet'):
            return kwargs.get('packet')

        if kwargs.get('raw'):
            pkt_raw = aprslib.parse(kwargs.get('raw'))
            pkt = core.factory(pkt_raw)
            return pkt

    def stats(self, serializable: bool = False) -> dict:
        return {
            'driver': self.__class__.__name__,
            'is_alive': self.is_alive(),
            'transport': 'fake',
        }



import unittest
from unittest import mock

from aprsd.client.drivers.registry import DriverRegistry

from ..mock_client_driver import MockClientDriver


class TestDriverRegistry(unittest.TestCase):
    """Unit tests for the DriverRegistry class."""

    def setUp(self):
        # Reset the singleton instance before each test
        DriverRegistry._singleton_instances = {}
        self.registry = DriverRegistry()
        self.registry.drivers = []

        # Mock APRSISDriver completely
        self.aprsis_patcher = mock.patch('aprsd.client.drivers.aprsis.APRSISDriver')
        mock_aprsis_class = self.aprsis_patcher.start()
        mock_aprsis_class.is_enabled.return_value = False
        mock_aprsis_class.is_configured.return_value = False

        # Mock the instance methods as well
        mock_instance = mock_aprsis_class.return_value
        mock_instance.is_enabled.return_value = False
        mock_instance.is_configured.return_value = False

        # Mock CONF to prevent password check
        self.conf_patcher = mock.patch('aprsd.client.drivers.aprsis.CONF')
        mock_conf = self.conf_patcher.start()
        mock_conf.aprs_network.password = 'dummy'
        mock_conf.aprs_network.login = 'dummy'

    def tearDown(self):
        # Reset the singleton instance after each test
        DriverRegistry().drivers = []
        self.aprsis_patcher.stop()
        self.conf_patcher.stop()

    def test_get_driver_with_valid_driver(self):
        """Test getting an enabled and configured driver."""
        # Add an enabled and configured driver
        driver = MockClientDriver
        driver.is_enabled = mock.MagicMock(return_value=True)
        driver.is_configured = mock.MagicMock(return_value=True)
        self.registry.register(MockClientDriver)

        # Get the driver
        result = self.registry.get_driver()
        print(result)
        self.assertTrue(isinstance(result, MockClientDriver))

    def test_get_driver_with_disabled_driver(self):
        """Test getting a driver when only disabled drivers exist."""
        driver = MockClientDriver
        driver.is_enabled = mock.MagicMock(return_value=False)
        driver.is_configured = mock.MagicMock(return_value=False)
        self.registry.register(driver)

        with self.assertRaises(ValueError) as context:
            self.registry.get_driver()
        self.assertIn('No enabled driver found', str(context.exception))

    def test_get_driver_with_unconfigured_driver(self):
        """Test getting a driver when only unconfigured drivers exist."""
        driver = MockClientDriver
        driver.is_enabled = mock.MagicMock(return_value=True)
        driver.is_configured = mock.MagicMock(return_value=False)
        self.registry.register(driver)

        with self.assertRaises(ValueError) as context:
            self.registry.get_driver()
        self.assertIn('No enabled driver found', str(context.exception))

    def test_get_driver_with_no_drivers(self):
        """Test getting a driver when no drivers exist."""
        # Try to get a driver
        with self.assertRaises(ValueError) as context:
            self.registry.get_driver()
        self.assertIn('No enabled driver found', str(context.exception))

    def test_get_driver_with_multiple_drivers(self):
        """Test getting a driver when multiple valid drivers exist."""
        # Add multiple drivers
        driver1 = MockClientDriver
        driver1.is_enabled = mock.MagicMock(return_value=True)
        driver1.is_configured = mock.MagicMock(return_value=True)
        driver2 = MockClientDriver
        self.registry.register(driver1)
        self.registry.register(driver2)

        # Get the driver - should return the first one
        result = self.registry.get_driver()
        # We can only check that it's a MockDriver instance
        self.assertTrue(isinstance(result, MockClientDriver))


if __name__ == '__main__':
    unittest.main()



from pathlib import Path

from oslo_config import cfg

home = str(Path.home())
DEFAULT_CONFIG_DIR = f'{home}/.config/aprsd/'
APRSD_DEFAULT_MAGIC_WORD = 'CHANGEME!!!'

watch_list_group = cfg.OptGroup(
    name='watch_list',
    title='Watch List settings',
)

registry_group = cfg.OptGroup(
    name='aprs_registry',
    title='APRS Registry settings',
)

aprsd_opts = [
    cfg.StrOpt(
        'callsign',
        default='NOCALL',
        help='Callsign to use for messages sent by APRSD',
    ),
    cfg.BoolOpt(
        'enable_save',
        default=True,
        help='Enable saving of watch list, packet tracker between restarts.',
    ),
    cfg.StrOpt(
        'save_location',
        default=DEFAULT_CONFIG_DIR,
        help='Save location for packet tracking files.',
    ),
    cfg.BoolOpt(
        'trace_enabled',
        default=False,
        help='Enable code tracing',
    ),
    cfg.StrOpt(
        'units',
        default='imperial',
        help='Units for display, imperial or metric',
    ),
    cfg.IntOpt(
        'ack_rate_limit_period',
        default=1,
        help='The wait period in seconds per Ack packet being sent.'
        '1 means 1 ack packet per second allowed.'
        '2 means 1 pack packet every 2 seconds allowed',
    ),
    cfg.IntOpt(
        'msg_rate_limit_period',
        default=2,
        help='Wait period in seconds per non AckPacket being sent.'
        '2 means 1 packet every 2 seconds allowed.'
        '5 means 1 pack packet every 5 seconds allowed',
    ),
    cfg.IntOpt(
        'packet_dupe_timeout',
        default=300,
        help='The number of seconds before a packet is not considered a duplicate.',
    ),
    cfg.BoolOpt(
        'enable_beacon',
        default=False,
        help='Enable sending of a GPS Beacon packet to locate this service. '
        'Requires latitude and longitude to be set.',
    ),
    cfg.IntOpt(
        'beacon_interval',
        default=1800,
        help='The number of seconds between beacon packets.',
    ),
    cfg.StrOpt(
        'beacon_symbol',
        default='/',
        help='The symbol to use for the GPS Beacon packet. See: http://www.aprs.net/vm/DOS/SYMBOLS.HTM',
    ),
    cfg.StrOpt(
        'latitude',
        default=None,
        help='Latitude for the GPS Beacon button.  If not set, the button will not be enabled.',
    ),
    cfg.StrOpt(
        'longitude',
        default=None,
        help='Longitude for the GPS Beacon button.  If not set, the button will not be enabled.',
    ),
    cfg.StrOpt(
        'log_packet_format',
        choices=['compact', 'multiline', 'both'],
        default='compact',
        help="When logging packets 'compact' will use a single line formatted for each packet."
        "'multiline' will use multiple lines for each packet and is the traditional format."
        'both will log both compact and multiline.',
    ),
    cfg.IntOpt(
        'default_packet_send_count',
        default=3,
        help='The number of times to send a non ack packet before giving up.',
    ),
    cfg.IntOpt(
        'default_ack_send_count',
        default=3,
        help='The number of times to send an ack packet in response to recieving a packet.',
    ),
    cfg.IntOpt(
        'packet_list_maxlen',
        default=100,
        help='The maximum number of packets to store in the packet list.',
    ),
    cfg.IntOpt(
        'packet_list_stats_maxlen',
        default=20,
        help='The maximum number of packets to send in the stats dict for admin ui. -1 means no max.',
    ),
    cfg.BoolOpt(
        'enable_seen_list',
        default=True,
        help='Enable the Callsign seen list tracking feature.  This allows aprsd to keep track of '
        'callsigns that have been seen and when they were last seen.',
    ),
    cfg.BoolOpt(
        'enable_packet_logging',
        default=True,
        help='Set this to False, to disable logging of packets to the log file.',
    ),
    cfg.BoolOpt(
        'load_help_plugin',
        default=True,
        help='Set this to False to disable the help plugin.',
    ),
    cfg.BoolOpt(
        'enable_sending_ack_packets',
        default=True,
        help='Set this to False, to disable sending of ack packets. This will entirely stop'
        'APRSD from sending ack packets.',
    ),
    cfg.BoolOpt(
        'is_digipi',
        default=False,
        help='Set this to True, if APRSD is running on a Digipi.'
        'This is useful for changing the behavior of APRSD to work with Digipi.',
    ),
]

watch_list_opts = [
    cfg.BoolOpt(
        'enabled',
        default=False,
        help='Enable the watch list feature.  Still have to enable '
        'the correct plugin.  Built-in plugin to use is '
        'aprsd.plugins.notify.NotifyPlugin',
    ),
    cfg.ListOpt(
        'callsigns',
        help='Callsigns to watch for messsages',
    ),
    cfg.StrOpt(
        'alert_callsign',
        help='The Ham Callsign to send messages to for watch list alerts.',
    ),
    cfg.IntOpt(
        'packet_keep_count',
        default=10,
        help='The number of packets to store.',
    ),
    cfg.IntOpt(
        'alert_time_seconds',
        default=3600,
        help='Time to wait before alert is sent on new message for users in callsigns.',
    ),
]


enabled_plugins_opts = [
    cfg.ListOpt(
        'enabled_plugins',
        default=[
            'aprsd.plugins.fortune.FortunePlugin',
            'aprsd.plugins.location.LocationPlugin',
            'aprsd.plugins.ping.PingPlugin',
            'aprsd.plugins.time.TimePlugin',
            'aprsd.plugins.weather.OWMWeatherPlugin',
            'aprsd.plugins.version.VersionPlugin',
            'aprsd.plugins.notify.NotifySeenPlugin',
        ],
        help='Comma separated list of enabled plugins for APRSD.'
        'To enable installed external plugins add them here.'
        'The full python path to the class name must be used',
    ),
]

registry_opts = [
    cfg.BoolOpt(
        'enabled',
        default=False,
        help='Enable sending aprs registry information.  This will let the '
        "APRS registry know about your service and it's uptime.  "
        'No personal information is sent, just the callsign, uptime and description. '
        'The service callsign is the callsign set in [DEFAULT] section.',
    ),
    cfg.StrOpt(
        'description',
        default=None,
        help='Description of the service to send to the APRS registry. '
        'This is what will show up in the APRS registry.'
        'If not set, the description will be the same as the callsign.',
    ),
    cfg.StrOpt(
        'registry_url',
        default='https://aprs.hemna.com/api/v1/registry',
        help='The APRS registry domain name to send the information to.',
    ),
    cfg.StrOpt(
        'service_website',
        default=None,
        help='The website for your APRS service to send to the APRS registry.',
    ),
    cfg.IntOpt(
        'frequency_seconds',
        default=3600,
        help='The frequency in seconds to send the APRS registry information.',
    ),
]


def register_opts(config):
    config.register_opts(aprsd_opts)
    config.register_opts(enabled_plugins_opts)
    config.register_group(watch_list_group)
    config.register_opts(watch_list_opts, group=watch_list_group)
    config.register_group(registry_group)
    config.register_opts(registry_opts, group=registry_group)


def list_opts():
    return {
        'DEFAULT': (aprsd_opts + enabled_plugins_opts),
        watch_list_group.name: watch_list_opts,
        registry_group.name: registry_opts,
    }



#!/usr/bin/env python3

"""Removes temporary Sphinx build artifacts to ensure a clean build.

This is needed if the Python source being documented changes significantly. Old sphinx-apidoc
RST files can be left behind.
"""

from pathlib import Path
import shutil


def main() -> None:
    docs_dir = Path(__file__).resolve().parent
    for folder in ("_build", "apidoc"):
        delete_dir = docs_dir / folder
        if delete_dir.exists():
            shutil.rmtree(delete_dir)


if __name__ == "__main__":
    main()



import datetime
import unittest
from unittest import mock

from aprslib.exceptions import LoginError

from aprsd import exception
from aprsd.client.drivers.aprsis import APRSISDriver
from aprsd.client.drivers.registry import ClientDriver
from aprsd.packets import core


class TestAPRSISDriver(unittest.TestCase):
    """Unit tests for the APRSISDriver class."""

    def setUp(self):
        # Mock configuration
        self.conf_patcher = mock.patch('aprsd.client.drivers.aprsis.CONF')
        self.mock_conf = self.conf_patcher.start()

        # Configure APRS-IS settings
        self.mock_conf.aprs_network.enabled = True
        self.mock_conf.aprs_network.login = 'TEST'
        self.mock_conf.aprs_network.password = '12345'
        self.mock_conf.aprs_network.host = 'rotate.aprs.net'
        self.mock_conf.aprs_network.port = 14580

        # Mock APRS Lib Client
        self.aprslib_patcher = mock.patch('aprsd.client.drivers.aprsis.APRSLibClient')
        self.mock_aprslib = self.aprslib_patcher.start()
        self.mock_client = mock.MagicMock()
        self.mock_aprslib.return_value = self.mock_client

        # Create an instance of the driver
        self.driver = APRSISDriver()
        self.driver.connected = False

    def tearDown(self):
        self.conf_patcher.stop()
        self.aprslib_patcher.stop()
        self.driver._client = None
        self.driver = None

    def test_implements_client_driver_protocol(self):
        """Test that APRSISDriver implements the ClientDriver Protocol."""
        # Verify the instance is recognized as implementing the Protocol
        self.assertIsInstance(self.driver, ClientDriver)

        # Verify all required methods are present with correct signatures
        required_methods = [
            'is_enabled',
            'is_configured',
            'is_alive',
            'close',
            'send',
            'setup_connection',
            'set_filter',
            'login_success',
            'login_failure',
            'consumer',
            'decode_packet',
            'stats',
        ]

        for method_name in required_methods:
            self.assertTrue(
                hasattr(self.driver, method_name),
                f'Missing required method: {method_name}',
            )

    def test_init(self):
        """Test initialization sets default values."""
        self.assertIsInstance(self.driver.max_delta, datetime.timedelta)
        self.assertEqual(self.driver.max_delta, datetime.timedelta(minutes=2))
        self.assertFalse(self.driver.login_status['success'])
        self.assertIsNone(self.driver.login_status['message'])
        self.assertIsNone(self.driver._client)

    def test_is_enabled_true(self):
        """Test is_enabled returns True when APRS-IS is enabled."""
        self.mock_conf.aprs_network.enabled = True
        self.assertTrue(APRSISDriver.is_enabled())

    def test_is_enabled_false(self):
        """Test is_enabled returns False when APRS-IS is disabled."""
        self.mock_conf.aprs_network.enabled = False
        self.assertFalse(APRSISDriver.is_enabled())

    def test_is_enabled_key_error(self):
        """Test is_enabled returns False when enabled flag doesn't exist."""
        self.mock_conf.aprs_network = mock.MagicMock()
        type(self.mock_conf.aprs_network).enabled = mock.PropertyMock(
            side_effect=KeyError
        )
        self.assertFalse(APRSISDriver.is_enabled())

    def test_is_configured_true(self):
        """Test is_configured returns True when properly configured."""
        with mock.patch.object(APRSISDriver, 'is_enabled', return_value=True):
            self.mock_conf.aprs_network.login = 'TEST'
            self.mock_conf.aprs_network.password = '12345'
            self.mock_conf.aprs_network.host = 'rotate.aprs.net'

            self.assertTrue(APRSISDriver.is_configured())

    def test_is_configured_no_login(self):
        """Test is_configured raises exception when login not set."""
        with mock.patch.object(APRSISDriver, 'is_enabled', return_value=True):
            self.mock_conf.aprs_network.login = None

            with self.assertRaises(exception.MissingConfigOptionException):
                APRSISDriver.is_configured()

    def test_is_configured_no_password(self):
        """Test is_configured raises exception when password not set."""
        with mock.patch.object(APRSISDriver, 'is_enabled', return_value=True):
            self.mock_conf.aprs_network.login = 'TEST'
            self.mock_conf.aprs_network.password = None

            with self.assertRaises(exception.MissingConfigOptionException):
                APRSISDriver.is_configured()

    def test_is_configured_no_host(self):
        """Test is_configured raises exception when host not set."""
        with mock.patch.object(APRSISDriver, 'is_enabled', return_value=True):
            self.mock_conf.aprs_network.login = 'TEST'
            self.mock_conf.aprs_network.password = '12345'
            self.mock_conf.aprs_network.host = None

            with self.assertRaises(exception.MissingConfigOptionException):
                APRSISDriver.is_configured()

    def test_is_configured_disabled(self):
        """Test is_configured returns True when not enabled."""
        with mock.patch.object(APRSISDriver, 'is_enabled', return_value=False):
            self.assertTrue(APRSISDriver.is_configured())

    def test_is_alive_no_client(self):
        """Test is_alive returns False when no client."""
        self.driver._client = None
        self.assertFalse(self.driver.is_alive)

    def test_is_alive_true(self):
        """Test is_alive returns True when client is alive and connection is not stale."""
        self.driver._client = self.mock_client
        self.mock_client.is_alive.return_value = True

        with mock.patch.object(self.driver, '_is_stale_connection', return_value=False):
            self.assertTrue(self.driver.is_alive)

    def test_is_alive_client_not_alive(self):
        """Test is_alive returns False when client is not alive."""
        self.driver._client = self.mock_client
        self.mock_client.is_alive.return_value = False

        with mock.patch.object(self.driver, '_is_stale_connection', return_value=False):
            self.assertFalse(self.driver.is_alive)

    def test_is_alive_stale_connection(self):
        """Test is_alive returns False when connection is stale."""
        self.driver._client = self.mock_client
        self.mock_client.is_alive.return_value = True

        with mock.patch.object(self.driver, '_is_stale_connection', return_value=True):
            self.assertFalse(self.driver.is_alive)

    def test_close(self):
        """Test close method stops and closes the client."""
        self.driver._client = self.mock_client

        self.driver.close()

        self.mock_client.stop.assert_called_once()
        self.mock_client.close.assert_called_once()

    def test_close_no_client(self):
        """Test close method handles no client gracefully."""
        self.driver._client = None

        # Should not raise exception
        self.driver.close()

    def test_send(self):
        """Test send passes packet to client."""
        self.driver._client = self.mock_client
        mock_packet = mock.MagicMock(spec=core.Packet)

        self.driver.send(mock_packet)

        self.mock_client.send.assert_called_once_with(mock_packet)

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    def test_setup_connection_success(self, mock_log):
        """Test setup_connection successfully connects."""
        # Configure successful connection
        self.mock_client.server_string = 'Test APRS-IS Server'

        self.driver.setup_connection()

        # Check client created with correct parameters
        self.mock_aprslib.assert_called_once_with(
            self.mock_conf.aprs_network.login,
            passwd=self.mock_conf.aprs_network.password,
            host=self.mock_conf.aprs_network.host,
            port=self.mock_conf.aprs_network.port,
        )

        # Check logger set and connection initialized
        self.assertEqual(self.mock_client.logger, mock_log)
        self.mock_client.connect.assert_called_once()

        # Check status updated
        self.assertTrue(self.driver.connected)
        self.assertTrue(self.driver.login_status['success'])
        self.assertEqual(self.driver.login_status['message'], 'Test APRS-IS Server')

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    @mock.patch('aprsd.client.drivers.aprsis.time.sleep')
    def test_setup_connection_login_error(self, mock_sleep, mock_log):
        """Test setup_connection handles login error."""
        # Configure login error
        login_error = LoginError('Bad login')
        login_error.message = 'Invalid login credentials'
        self.mock_client.connect.side_effect = login_error

        self.driver.setup_connection()

        # Check error logged
        mock_log.error.assert_any_call("Failed to login to APRS-IS Server 'Bad login'")
        mock_log.error.assert_any_call('Invalid login credentials')

        # Check status updated
        self.assertFalse(self.driver.connected)
        self.assertFalse(self.driver.login_status['success'])
        self.assertEqual(
            self.driver.login_status['message'], 'Invalid login credentials'
        )

        # Check backoff used
        mock_sleep.assert_called()

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    @mock.patch('aprsd.client.drivers.aprsis.time.sleep')
    def test_setup_connection_general_error(self, mock_sleep, mock_log):
        """Test setup_connection handles general error."""
        # Configure general exception
        error_message = 'Connection error'
        error = Exception(error_message)
        # Standard exceptions don't have a message attribute
        self.mock_client.connect.side_effect = error

        self.driver.setup_connection()

        # Check error logged
        mock_log.error.assert_any_call(
            f"Unable to connect to APRS-IS server. '{error_message}' "
        )

        # Check status updated
        self.assertFalse(self.driver.connected)
        self.assertFalse(self.driver.login_status['success'])

        # Check login message contains the error message (more flexible than exact equality)
        self.assertIn(error_message, self.driver.login_status['message'])

        # Check backoff used
        mock_sleep.assert_called()

    def test_set_filter(self):
        """Test set_filter passes filter to client."""
        self.driver._client = self.mock_client
        test_filter = 'm/50'

        self.driver.set_filter(test_filter)

        self.mock_client.set_filter.assert_called_once_with(test_filter)

    def test_login_success(self):
        """Test login_success returns login status."""
        self.driver.login_status['success'] = True
        self.assertTrue(self.driver.login_success())

        self.driver.login_status['success'] = False
        self.assertFalse(self.driver.login_success())

    def test_login_failure(self):
        """Test login_failure returns error message."""
        self.driver.login_status['message'] = None
        self.assertIsNone(self.driver.login_failure())

        self.driver.login_status['message'] = 'Test error'
        self.assertEqual(self.driver.login_failure(), 'Test error')

    def test_filter_property(self):
        """Test filter property returns client filter."""
        self.driver._client = self.mock_client
        test_filter = 'm/50'
        self.mock_client.filter = test_filter

        self.assertEqual(self.driver.filter, test_filter)

    def test_server_string_property(self):
        """Test server_string property returns client server string."""
        self.driver._client = self.mock_client
        test_string = 'Test APRS-IS Server'
        self.mock_client.server_string = test_string

        self.assertEqual(self.driver.server_string, test_string)

    def test_keepalive_property(self):
        """Test keepalive property returns client keepalive."""
        self.driver._client = self.mock_client
        test_time = datetime.datetime.now()
        self.mock_client.aprsd_keepalive = test_time

        self.assertEqual(self.driver.keepalive, test_time)

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    def test_is_stale_connection_true(self, mock_log):
        """Test _is_stale_connection returns True when connection is stale."""
        self.driver._client = self.mock_client
        # Set keepalive to 3 minutes ago (exceeds max_delta of 2 minutes)
        self.mock_client.aprsd_keepalive = datetime.datetime.now() - datetime.timedelta(
            minutes=3
        )

        result = self.driver._is_stale_connection()

        self.assertTrue(result)
        mock_log.warning.assert_called_once()

    def test_is_stale_connection_false(self):
        """Test _is_stale_connection returns False when connection is not stale."""
        self.driver._client = self.mock_client
        # Set keepalive to 1 minute ago (within max_delta of 2 minutes)
        self.mock_client.aprsd_keepalive = datetime.datetime.now() - datetime.timedelta(
            minutes=1
        )

        result = self.driver._is_stale_connection()

        self.assertFalse(result)

    def test_transport(self):
        """Test transport returns appropriate transport type."""
        self.assertEqual(APRSISDriver.transport(), 'aprsis')

    def test_decode_packet(self):
        """Test decode_packet uses core.factory."""
        with mock.patch('aprsd.client.drivers.aprsis.core.factory') as mock_factory:
            raw_packet = {'from': 'TEST', 'to': 'APRS'}
            self.driver.decode_packet(raw_packet)
            mock_factory.assert_called_once_with(raw_packet)

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    def test_consumer_success(self, mock_log):
        """Test consumer forwards callback to client."""
        self.driver._client = self.mock_client
        self.driver.connected = True
        mock_callback = mock.MagicMock()

        self.driver.consumer(mock_callback, raw=True)

        self.mock_client.consumer.assert_called_once_with(
            mock_callback, blocking=False, immortal=False, raw=True
        )

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    def test_consumer_exception(self, mock_log):
        """Test consumer handles exceptions."""
        self.driver._client = self.mock_client
        self.driver.connected = True
        mock_callback = mock.MagicMock()
        test_error = Exception('Test error')
        self.mock_client.consumer.side_effect = test_error

        with self.assertRaises(Exception):  # noqa: B017
            self.driver.consumer(mock_callback)

        mock_log.error.assert_called_with(test_error)

    @mock.patch('aprsd.client.drivers.aprsis.LOG')
    def test_consumer_no_client(self, mock_log):
        """Test consumer handles no client gracefully."""
        self.driver._client = None
        mock_callback = mock.MagicMock()

        self.driver.consumer(mock_callback)
        self.assertFalse(self.driver.connected)

    def test_stats_configured_with_client(self):
        """Test stats returns correct data when configured with client."""
        # Configure driver
        with mock.patch.object(self.driver, 'is_configured', return_value=True):
            self.driver._client = self.mock_client
            self.mock_client.aprsd_keepalive = datetime.datetime.now()
            self.mock_client.server_string = 'Test Server'
            self.mock_client.filter = 'm/50'

            stats = self.driver.stats()

            self.assertEqual(stats['connected'], True)
            self.assertEqual(stats['filter'], 'm/50')
            self.assertEqual(stats['server_string'], 'Test Server')
            self.assertEqual(stats['transport'], 'aprsis')

    def test_stats_serializable(self):
        """Test stats with serializable=True converts datetime to ISO format."""
        # Configure driver
        with mock.patch.object(self.driver, 'is_configured', return_value=True):
            self.driver._client = self.mock_client
            test_time = datetime.datetime.now()
            self.mock_client.aprsd_keepalive = test_time

            stats = self.driver.stats(serializable=True)

            # Check keepalive is a string in ISO format
            self.assertIsInstance(stats['connection_keepalive'], str)
            # Try parsing it to verify it's a valid ISO format
            try:
                datetime.datetime.fromisoformat(stats['connection_keepalive'])
            except ValueError:
                self.fail('keepalive is not in valid ISO format')

    def test_stats_no_client(self):
        """Test stats with no client."""
        with mock.patch.object(self.driver, 'is_configured', return_value=True):
            self.driver._client = None

            stats = self.driver.stats()

            self.assertEqual(stats['connection_keepalive'], 'None')
            self.assertEqual(stats['server_string'], 'None')

    def test_stats_not_configured(self):
        """Test stats when not configured returns empty dict."""
        with mock.patch.object(self.driver, 'is_configured', return_value=False):
            stats = self.driver.stats()
            self.assertEqual(stats, {})


if __name__ == '__main__':
    unittest.main()



class MissingConfigOptionException(Exception):
    """Missing a config option."""

    def __init__(self, config_option):
        self.message = f"Option '{config_option}' was not in config file"


class ConfigOptionBogusDefaultException(Exception):
    """Missing a config option."""

    def __init__(self, config_option, default_fail):
        self.message = (
            f"Config file option '{config_option}' needs to be "
            f"changed from provided default of '{default_fail}'"
        )


class APRSClientNotConfiguredException(Exception):
    """APRS client is not configured."""

    def __init__(self):
        self.message = 'APRS client is not configured.'



from unittest import mock

from oslo_config import cfg

from aprsd import conf  # noqa: F401
from aprsd.plugins import fortune as fortune_plugin

from .. import fake, test_plugin


CONF = cfg.CONF


class TestFortunePlugin(test_plugin.TestPlugin):
    @mock.patch("shutil.which")
    def test_fortune_fail(self, mock_which):
        mock_which.return_value = None
        fortune = fortune_plugin.FortunePlugin()
        expected = "FortunePlugin isn't enabled"
        packet = fake.fake_packet(message="fortune")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("subprocess.check_output")
    @mock.patch("shutil.which")
    def test_fortune_success(self, mock_which, mock_output):
        mock_which.return_value = "/usr/bin/games/fortune"
        mock_output.return_value = "Funny fortune"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        fortune = fortune_plugin.FortunePlugin()

        expected = "Funny fortune"
        packet = fake.fake_packet(message="fortune")
        actual = fortune.filter(packet)
        self.assertEqual(expected, actual)



import datetime
import logging

from oslo_config import cfg

from aprsd import utils
from aprsd.packets import core
from aprsd.utils import objectstore


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class WatchList(objectstore.ObjectStoreMixin):
    """Global watch list and info for callsigns."""

    _instance = None
    data = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        super().__init__()
        self._update_from_conf()

    def _update_from_conf(self, config=None):
        with self.lock:
            if CONF.watch_list.enabled and CONF.watch_list.callsigns:
                for callsign in CONF.watch_list.callsigns:
                    call = callsign.replace("*", "")
                    # FIXME(waboring) - we should fetch the last time we saw
                    # a beacon from a callsign or some other mechanism to find
                    # last time a message was seen by aprs-is.  For now this
                    # is all we can do.
                    if call not in self.data:
                        self.data[call] = {
                            "last": None,
                            "packet": None,
                        }

    def stats(self, serializable=False) -> dict:
        stats = {}
        with self.lock:
            for callsign in self.data:
                stats[callsign] = {
                    "last": self.data[callsign]["last"],
                    "packet": self.data[callsign]["packet"],
                    "age": self.age(callsign),
                    "old": self.is_old(callsign),
                }
        return stats

    def is_enabled(self):
        return CONF.watch_list.enabled

    def callsign_in_watchlist(self, callsign):
        with self.lock:
            return callsign in self.data

    def rx(self, packet: type[core.Packet]) -> None:
        """Track when we got a packet from the network."""
        callsign = packet.from_call

        if self.callsign_in_watchlist(callsign):
            with self.lock:
                self.data[callsign]["last"] = datetime.datetime.now()
                self.data[callsign]["packet"] = packet

    def tx(self, packet: type[core.Packet]) -> None:
        """We don't care about TX packets."""

    def last_seen(self, callsign):
        with self.lock:
            if self.callsign_in_watchlist(callsign):
                return self.data[callsign]["last"]

    def age(self, callsign):
        now = datetime.datetime.now()
        last_seen_time = self.last_seen(callsign)
        if last_seen_time:
            return str(now - last_seen_time)
        else:
            return None

    def max_delta(self, seconds=None):
        if not seconds:
            seconds = CONF.watch_list.alert_time_seconds
        max_timeout = {"seconds": seconds}
        return datetime.timedelta(**max_timeout)

    def is_old(self, callsign, seconds=None):
        """Watch list callsign last seen is old compared to now?

        This tests to see if the last time we saw a callsign packet,
        if that is older than the allowed timeout in the config.

        We put this here so any notification plugin can use this
        same test.
        """
        if not self.callsign_in_watchlist(callsign):
            return False

        age = self.age(callsign)
        if age:
            delta = utils.parse_delta_str(age)
            d = datetime.timedelta(**delta)

            max_delta = self.max_delta(seconds=seconds)

            if d > max_delta:
                return True
            else:
                return False
        else:
            return False



import click
import click.shell_completion

from aprsd.main import cli

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@cli.command()
@click.argument(
    "shell", type=click.Choice(list(click.shell_completion._available_shells))
)
def completion(shell):
    """Show the shell completion code"""
    from click.utils import _detect_program_name

    cls = click.shell_completion.get_completion_class(shell)
    prog_name = _detect_program_name()
    complete_var = f"_{prog_name}_COMPLETE".replace("-", "_").upper()
    print(cls(cli, {}, prog_name, complete_var).source())
    print(
        "# Add the following line to your shell configuration file to have aprsd command line completion"
    )
    print("# but remove the leading '#' character.")
    print(f'# eval "$(aprsd completion {shell})"')



"""
The options for log setup
"""

from oslo_config import cfg

DEFAULT_LOGIN = "NOCALL"

aprs_group = cfg.OptGroup(
    name="aprs_network",
    title="APRS-IS Network settings",
)

kiss_serial_group = cfg.OptGroup(
    name="kiss_serial",
    title="KISS Serial device connection",
)

kiss_tcp_group = cfg.OptGroup(
    name="kiss_tcp",
    title="KISS TCP/IP Device connection",
)

fake_client_group = cfg.OptGroup(
    name="fake_client",
    title="Fake Client settings",
)
aprs_opts = [
    cfg.BoolOpt(
        "enabled",
        default=True,
        help="Set enabled to False if there is no internet connectivity."
        "This is useful for a direwolf KISS aprs connection only.",
    ),
    cfg.StrOpt(
        "login",
        default=DEFAULT_LOGIN,
        help="APRS Username",
    ),
    cfg.StrOpt(
        "password",
        secret=True,
        help="APRS Password "
        "Get the passcode for your callsign here: "
        "https://apps.magicbug.co.uk/passcode",
    ),
    cfg.HostAddressOpt(
        "host",
        default="noam.aprs2.net",
        help="The APRS-IS hostname",
    ),
    cfg.PortOpt(
        "port",
        default=14580,
        help="APRS-IS port",
    ),
]

kiss_serial_opts = [
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable Serial KISS interface connection.",
    ),
    cfg.StrOpt(
        "device",
        help="Serial Device file to use.  /dev/ttyS0",
    ),
    cfg.IntOpt(
        "baudrate",
        default=9600,
        help="The Serial device baud rate for communication",
    ),
    cfg.ListOpt(
        "path",
        default=["WIDE1-1", "WIDE2-1"],
        help="The APRS path to use for wide area coverage.",
    ),
]

kiss_tcp_opts = [
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable Serial KISS interface connection.",
    ),
    cfg.HostAddressOpt(
        "host",
        help="The KISS TCP Host to connect to.",
    ),
    cfg.PortOpt(
        "port",
        default=8001,
        help="The KISS TCP/IP network port",
    ),
    cfg.ListOpt(
        "path",
        default=["WIDE1-1", "WIDE2-1"],
        help="The APRS path to use for wide area coverage.",
    ),
]

fake_client_opts = [
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable fake client connection.",
    ),
]


def register_opts(config):
    config.register_group(aprs_group)
    config.register_opts(aprs_opts, group=aprs_group)
    config.register_group(kiss_serial_group)
    config.register_group(kiss_tcp_group)
    config.register_opts(kiss_serial_opts, group=kiss_serial_group)
    config.register_opts(kiss_tcp_opts, group=kiss_tcp_group)

    config.register_group(fake_client_group)
    config.register_opts(fake_client_opts, group=fake_client_group)


def list_opts():
    return {
        aprs_group.name: aprs_opts,
        kiss_serial_group.name: kiss_serial_opts,
        kiss_tcp_group.name: kiss_tcp_opts,
        fake_client_group.name: fake_client_opts,
    }



import logging
from typing import Callable, Protocol, runtime_checkable

from aprsd.packets import core
from aprsd.utils import singleton


LOG = logging.getLogger("APRSD")


@runtime_checkable
class PacketMonitor(Protocol):
    """Protocol for Monitoring packets in some way."""

    def rx(self, packet: type[core.Packet]) -> None:
        """When we get a packet from the network."""
        ...

    def tx(self, packet: type[core.Packet]) -> None:
        """When we send a packet out the network."""
        ...

    def flush(self) -> None:
        """Flush out any data."""
        ...

    def load(self) -> None:
        """Load any data."""
        ...


@singleton
class PacketCollector:
    def __init__(self):
        self.monitors: list[Callable] = []

    def register(self, monitor: Callable) -> None:
        if not isinstance(monitor, PacketMonitor):
            raise TypeError(f"Monitor {monitor} is not a PacketMonitor")
        self.monitors.append(monitor)

    def unregister(self, monitor: Callable) -> None:
        if not isinstance(monitor, PacketMonitor):
            raise TypeError(f"Monitor {monitor} is not a PacketMonitor")
        self.monitors.remove(monitor)

    def rx(self, packet: type[core.Packet]) -> None:
        for name in self.monitors:
            cls = name()
            try:
                cls.rx(packet)
            except Exception as e:
                LOG.error(f"Error in monitor {name} (rx): {e}")

    def tx(self, packet: type[core.Packet]) -> None:
        for name in self.monitors:
            cls = name()
            try:
                cls.tx(packet)
            except Exception as e:
                LOG.error(f"Error in monitor {name} (tx): {e}")

    def flush(self):
        """Call flush on the objects. This is used to flush out any data."""
        for name in self.monitors:
            cls = name()
            try:
                cls.flush()
            except Exception as e:
                LOG.error(f"Error in monitor {name} (flush): {e}")

    def load(self):
        """Call load on the objects. This is used to load any data."""
        for name in self.monitors:
            cls = name()
            try:
                cls.load()
            except Exception as e:
                LOG.error(f"Error in monitor {name} (load): {e}")



import logging
import time

from oslo_config import cfg

from aprsd.stats import collector
from aprsd.threads import APRSDThread
from aprsd.utils import objectstore

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class StatsStore(objectstore.ObjectStoreMixin):
    """Container to save the stats from the collector."""

    def add(self, stats: dict):
        with self.lock:
            self.data = stats


class APRSDStatsStoreThread(APRSDThread):
    """Save APRSD Stats to disk periodically."""

    # how often in seconds to write the file
    save_interval = 10

    def __init__(self):
        super().__init__('StatsStore')

    def loop(self):
        if self.loop_count % self.save_interval == 0:
            stats = collector.Collector().collect()
            ss = StatsStore()
            ss.add(stats)
            ss.save()

        time.sleep(1)
        return True



"""
APRSD KISS Client Driver using native KISS implementation.

This module provides a KISS client driver for APRSD using the new
non-asyncio KISSInterface implementation.
"""

import datetime
import logging
import select
import socket
from typing import Any, Callable, Dict

import aprslib
from ax253 import frame as ax25frame
from kiss import constants as kiss_constants
from kiss import util as kissutil
from kiss.kiss import Command
from oslo_config import cfg

from aprsd import (  # noqa
    client,
    conf,  # noqa
    exception,
)
from aprsd.packets import core
from aprsd.utils import trace

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


def handle_fend(buffer: bytes, strip_df_start: bool = True) -> bytes:
    """
    Handle FEND (end of frame) encountered in a KISS data stream.

    :param buffer: the buffer containing the frame
    :param strip_df_start: remove leading null byte (DATA_FRAME opcode)
    :return: the bytes of the frame without escape characters or frame
             end markers (FEND)
    """
    frame = kissutil.recover_special_codes(kissutil.strip_nmea(bytes(buffer)))
    if strip_df_start:
        frame = kissutil.strip_df_start(frame)
    LOG.warning(f'handle_fend {" ".join(f"{b:02X}" for b in bytes(frame))}')
    return bytes(frame)


class TCPKISSDriver(metaclass=trace.TraceWrapperMetaclass):
    # class TCPKISSDriver:
    """APRSD client driver for TCP KISS connections."""

    _instance = None

    # Class level attributes required by Client protocol
    packets_received = 0
    packets_sent = 0
    last_packet_sent = None
    last_packet_received = None
    keepalive = None
    client_name = None
    socket = None
    # timeout in seconds
    select_timeout = 1
    path = None

    def __new__(cls, *args, **kwargs):
        """This magic turns this into a singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        """Initialize the KISS client.

        Args:
            client_name: Name of the client instance
        """
        super().__init__()
        self._connected = False
        self.keepalive = datetime.datetime.now()
        # This is initialized in setup_connection()
        self.socket = None

    @property
    def transport(self) -> str:
        return client.TRANSPORT_TCPKISS

    @staticmethod
    def is_enabled() -> bool:
        """Check if KISS is enabled in configuration.

        Returns:
            bool: True if either TCP is enabled
        """
        return CONF.kiss_tcp.enabled

    @staticmethod
    def is_configured():
        # Ensure that the config vars are correctly set
        if TCPKISSDriver.is_enabled():
            if not CONF.kiss_tcp.host:
                LOG.error('KISS TCP enabled, but no host is set.')
                raise exception.MissingConfigOptionException(
                    'kiss_tcp.host is not set.',
                )
            return True
        return False

    @property
    def is_alive(self) -> bool:
        """Check if the client is connected.

        Returns:
            bool: True if connected to KISS TNC, False otherwise
        """
        return self._connected

    def close(self):
        """Close the connection."""
        self._connected = False
        if self.socket:
            try:
                self.socket.close()
            except Exception as e:
                LOG.error(f'close: error closing socket: {e}')
                pass
        else:
            LOG.warning('close: socket not initialized. no reason to close.')

    def send(self, packet: core.Packet):
        """Send an APRS packet.

        Args:
            packet: APRS packet to send (Packet or Message object)

        Raises:
            Exception: If not connected or send fails
        """
        if not self.socket:
            raise Exception('KISS interface not initialized')

        payload = None
        path = self.path
        packet.prepare()
        payload = packet.payload.encode('US-ASCII')
        if packet.path:
            path = packet.path

        LOG.debug(
            f"KISS Send '{payload}' TO '{packet.to_call}' From "
            f"'{packet.from_call}' with PATH '{path}'",
        )
        frame = ax25frame.Frame.ui(
            destination='APZ100',
            # destination=packet.to_call,
            source=packet.from_call,
            path=path,
            info=payload,
        )

        # now escape the frame special characters
        frame_escaped = kissutil.escape_special_codes(bytes(frame))
        # and finally wrap the frame in KISS protocol
        command = Command.DATA_FRAME
        frame_kiss = b''.join(
            [kiss_constants.FEND, command.value, frame_escaped, kiss_constants.FEND]
        )
        self.socket.send(frame_kiss)
        # Update last packet sent time
        self.last_packet_sent = datetime.datetime.now()
        # Increment packets sent counter
        self.packets_sent += 1

    def setup_connection(self):
        """Set up the KISS interface."""
        if not self.is_enabled():
            LOG.error('KISS is not enabled in configuration')
            return

        if self._connected:
            LOG.warning('KISS interface already connected')
            return

        try:
            # Configure for TCP KISS
            if self.is_enabled():
                LOG.info(
                    f'KISS TCP Connection to {CONF.kiss_tcp.host}:{CONF.kiss_tcp.port}'
                )
                self.path = CONF.kiss_tcp.path
                self.connect()
            if self._connected:
                LOG.info('KISS interface initialized')
            else:
                LOG.error('Failed to connect to KISS interface')

        except Exception as ex:
            LOG.error('Failed to initialize KISS interface')
            LOG.exception(ex)
            self._connected = False

    def set_filter(self, filter_text: str):
        """Set packet filter (not implemented for KISS).

        Args:
            filter_text: Filter specification (ignored for KISS)
        """
        # KISS doesn't support filtering at the TNC level
        pass

    @property
    def filter(self) -> str:
        """Get packet filter (not implemented for KISS).
        Returns:
            str: Empty string (not implemented for KISS)
        """
        return ''

    def login_success(self) -> bool:
        """There is no login for KISS."""
        if not self._connected:
            return False
        return True

    def login_failure(self) -> str:
        """There is no login for KISS."""
        return 'Login successful'

    def consumer(self, callback: Callable, raw: bool = False):
        """Start consuming frames with the given callback.

        Args:
            callback: Function to call with received packets

        Raises:
            Exception: If not connected to KISS TNC
        """
        # Ensure connection
        if not self._connected:
            return

        # Read frame
        frame = self.read_frame()
        if frame:
            LOG.info(f'GOT FRAME: {frame} calling {callback}')
            kwargs = {
                'frame': frame,
            }
            callback(**kwargs)

    def decode_packet(self, *args, **kwargs) -> core.Packet:
        """Decode a packet from an AX.25 frame.

        Args:
            frame: Received AX.25 frame
        """
        frame = kwargs.get('frame')
        if not frame:
            LOG.warning('No frame received to decode?!?!')
            return None

        try:
            aprslib_frame = aprslib.parse(str(frame))
            packet = core.factory(aprslib_frame)
            if isinstance(packet, core.ThirdPartyPacket):
                return packet.subpacket
            else:
                return packet
        except Exception as e:
            LOG.error(f'Error decoding packet: {e}')
            return None

    def stats(self, serializable: bool = False) -> Dict[str, Any]:
        """Get client statistics.

        Returns:
            Dict containing client statistics
        """
        if serializable:
            keepalive = self.keepalive.isoformat()
            if self.last_packet_sent:
                last_packet_sent = self.last_packet_sent.isoformat()
            else:
                last_packet_sent = 'None'
            if self.last_packet_received:
                last_packet_received = self.last_packet_received.isoformat()
            else:
                last_packet_received = 'None'
        else:
            keepalive = self.keepalive
            last_packet_sent = self.last_packet_sent
            last_packet_received = self.last_packet_received

        stats = {
            'client': self.__class__.__name__,
            'transport': self.transport,
            'connected': self._connected,
            'path': self.path,
            'packets_sent': self.packets_sent,
            'packets_received': self.packets_received,
            'last_packet_sent': last_packet_sent,
            'last_packet_received': last_packet_received,
            'connection_keepalive': keepalive,
            'host': CONF.kiss_tcp.host,
            'port': CONF.kiss_tcp.port,
        }

        return stats

    def connect(self) -> bool:
        """Establish TCP connection to the KISS host.

        Returns:
            bool: True if connection successful, False otherwise
        """
        try:
            if self.socket:
                try:
                    self.socket.close()
                except Exception:
                    pass

            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.settimeout(5.0)  # 5 second timeout for connection
            self.socket.connect((CONF.kiss_tcp.host, CONF.kiss_tcp.port))
            self.socket.settimeout(0.1)  # Reset to shorter timeout for reads
            self._connected = True
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # MACOS doesn't have TCP_KEEPIDLE
            if hasattr(socket, 'TCP_KEEPIDLE'):
                self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 3)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
            return True

        except ConnectionError as e:
            LOG.error(
                f'Failed to connect to {CONF.kiss_tcp.host}:{CONF.kiss_tcp.port} - {str(e)}'
            )
            self._connected = False
            return False

        except Exception as e:
            LOG.error(
                f'Failed to connect to {CONF.kiss_tcp.host}:{CONF.kiss_tcp.port} - {str(e)}'
            )
            self._connected = False
            return False

    def fix_raw_frame(self, raw_frame: bytes) -> bytes:
        """Fix the raw frame by recalculating the FCS."""
        ax25_data = raw_frame[2:-1]  # Remove KISS markers
        return handle_fend(ax25_data)

    def read_frame(self, blocking=False):
        """
        Generator for complete lines, received from the server
        """
        if not self.socket:
            return None

        if not self._connected:
            return None

        try:
            self.socket.setblocking(0)
        except OSError as e:
            LOG.error(f'socket error when setblocking(0): {str(e)}')
            raise aprslib.ConnectionDrop('connection dropped') from e

        while self._connected:
            short_buf = b''

            try:
                readable, _, _ = select.select(
                    [self.socket],
                    [],
                    [],
                    self.select_timeout,
                )
                if not readable:
                    if not blocking:
                        break
                    else:
                        continue
            except Exception as e:
                # No need to log if we are not running.
                # this happens when the client is stopped/closed.
                LOG.error(f'Error in read loop: {e}')
                self._connected = False
                break

            try:
                short_buf = self.socket.recv(1024)
                # sock.recv returns empty if the connection drops
                if not short_buf:
                    if not blocking:
                        # We could just not be blocking, so empty is expected
                        continue
                    else:
                        self.logger.error('socket.recv(): returned empty')
                        raise aprslib.ConnectionDrop('connection dropped')

                raw_frame = self.fix_raw_frame(short_buf)
                return ax25frame.Frame.from_bytes(raw_frame)
            except OSError as e:
                # self.logger.error("socket error on recv(): %s" % str(e))
                if 'Resource temporarily unavailable' in str(e):
                    if not blocking:
                        if len(short_buf) == 0:
                            break
            except socket.timeout:
                continue
            except (KeyboardInterrupt, SystemExit):
                raise
            except ConnectionError:
                self.close()
                if not self.auto_reconnect:
                    raise
                else:
                    self.connect()
                    continue
            except StopIteration:
                break
            except IOError:
                LOG.error('IOError')
                break
            except Exception as e:
                LOG.error(f'Error in read loop: {e}')
                self._connected = False
                if not self.auto_reconnect:
                    break



from oslo_config import cfg

aprsfi_group = cfg.OptGroup(
    name="aprs_fi",
    title="APRS.FI website settings",
)
query_group = cfg.OptGroup(
    name="query_plugin",
    title="Options for the Query Plugin",
)
avwx_group = cfg.OptGroup(
    name="avwx_plugin",
    title="Options for the AVWXWeatherPlugin",
)
owm_wx_group = cfg.OptGroup(
    name="owm_weather_plugin",
    title="Options for the OWMWeatherPlugin",
)

aprsfi_opts = [
    cfg.StrOpt(
        "apiKey",
        help="Get the apiKey from your aprs.fi account here:" "http://aprs.fi/account",
    ),
]

owm_wx_opts = [
    cfg.StrOpt(
        "apiKey",
        help="OWMWeatherPlugin api key to OpenWeatherMap's API."
        "This plugin uses the openweathermap API to fetch"
        "location and weather information."
        "To use this plugin you need to get an openweathermap"
        "account and apikey."
        "https://home.openweathermap.org/api_keys",
    ),
]

avwx_opts = [
    cfg.StrOpt(
        "apiKey",
        help="avwx-api is an opensource project that has"
        "a hosted service here: https://avwx.rest/"
        "You can launch your own avwx-api in a container"
        "by cloning the githug repo here:"
        "https://github.com/avwx-rest/AVWX-API",
    ),
    cfg.StrOpt(
        "base_url",
        default="https://avwx.rest",
        help="The base url for the avwx API.  If you are hosting your own"
        "Here is where you change the url to point to yours.",
    ),
]


def register_opts(config):
    config.register_group(aprsfi_group)
    config.register_opts(aprsfi_opts, group=aprsfi_group)
    config.register_group(query_group)
    config.register_group(owm_wx_group)
    config.register_opts(owm_wx_opts, group=owm_wx_group)
    config.register_group(avwx_group)
    config.register_opts(avwx_opts, group=avwx_group)


def list_opts():
    return {
        aprsfi_group.name: aprsfi_opts,
        owm_wx_group.name: owm_wx_opts,
        avwx_group.name: avwx_opts,
    }



import logging
import signal
import sys

import click
from oslo_config import cfg

import aprsd
from aprsd import cli_helper, plugin, threads, utils
from aprsd import main as aprsd_main
from aprsd.client.client import APRSDClient
from aprsd.main import cli
from aprsd.packets import collector as packet_collector
from aprsd.packets import seen_list
from aprsd.threads import keepalive, registry, rx, service, tx
from aprsd.threads import stats as stats_thread

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


# main() ###
@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    '-f',
    '--flush',
    'flush',
    is_flag=True,
    show_default=True,
    default=False,
    help='Flush out all old aged messages on disk.',
)
@click.pass_context
@cli_helper.process_standard_options
def server(ctx, flush):
    """Start the aprsd server gateway process."""
    signal.signal(signal.SIGINT, aprsd_main.signal_handler)
    signal.signal(signal.SIGTERM, aprsd_main.signal_handler)

    service_threads = service.ServiceThreads()

    LOG.info(f'Python version: {sys.version}')
    LOG.info(f'APRSD Started version: {aprsd.__version__}')
    level, msg = utils._check_version()
    if level:
        LOG.warning(msg)
    else:
        LOG.info(msg)
    utils.package.log_installed_extensions_and_plugins()

    # Make sure we have 1 client transport enabled
    if not APRSDClient().is_enabled:
        LOG.error('No Clients are enabled in config.')
        sys.exit(-1)

    if not APRSDClient().is_configured:
        LOG.error('APRS client is not properly configured in config file.')
        sys.exit(-1)

    # Creates the client object
    LOG.info('Creating client connection')
    aprs_client = APRSDClient()
    LOG.info(aprs_client)
    if not aprs_client.login_success:
        # We failed to login, will just quit!
        msg = f'Login Failure: {aprs_client.login_failure}'
        LOG.error(msg)
        print(msg)
        sys.exit(-1)

    # Check to make sure the login worked.

    # Create the initial PM singleton and Register plugins
    # We register plugins first here so we can register each
    # plugins config options, so we can dump them all in the
    # log file output.
    LOG.info('Loading Plugin Manager and registering plugins')
    plugin_manager = plugin.PluginManager()
    plugin_manager.setup_plugins(load_help_plugin=CONF.load_help_plugin)

    # Dump all the config options now.
    CONF.log_opt_values(LOG, logging.DEBUG)
    message_plugins = plugin_manager.get_message_plugins()
    watchlist_plugins = plugin_manager.get_watchlist_plugins()
    LOG.info('Message Plugins enabled and running:')
    for p in message_plugins:
        LOG.info(p)
    LOG.info('Watchlist Plugins enabled and running:')
    for p in watchlist_plugins:
        LOG.info(p)

    if not CONF.enable_seen_list:
        # just deregister the class from the packet collector
        packet_collector.PacketCollector().unregister(seen_list.SeenList)

    # Now load the msgTrack from disk if any
    if flush:
        LOG.debug('Flushing All packet tracking objects.')
        packet_collector.PacketCollector().flush()
    else:
        # Try and load saved MsgTrack list
        LOG.debug('Loading saved packet tracking data.')
        packet_collector.PacketCollector().load()

    # Now start all the main processing threads.

    service_threads.register(keepalive.KeepAliveThread())
    service_threads.register(stats_thread.APRSDStatsStoreThread())
    service_threads.register(
        rx.APRSDRXThread(
            packet_queue=threads.packet_queue,
        ),
    )
    service_threads.register(
        rx.APRSDPluginProcessPacketThread(
            packet_queue=threads.packet_queue,
        ),
    )

    if CONF.enable_beacon:
        LOG.info('Beacon Enabled.  Starting Beacon thread.')
        service_threads.register(tx.BeaconSendThread())

    if CONF.aprs_registry.enabled:
        LOG.info('Registry Enabled.  Starting Registry thread.')
        service_threads.register(registry.APRSRegistryThread())

    service_threads.start()
    service_threads.join()

    return 0



import datetime
import logging

from oslo_config import cfg

from aprsd.packets import core
from aprsd.utils import objectstore


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class PacketTrack(objectstore.ObjectStoreMixin):
    """Class to keep track of outstanding text messages.

    This is a thread safe class that keeps track of active
    messages.

    When a message is asked to be sent, it is placed into this
    class via it's id.  The TextMessage class's send() method
    automatically adds itself to this class.  When the ack is
    recieved from the radio, the message object is removed from
    this class.
    """

    _instance = None
    _start_time = None

    data: dict = {}
    total_tracked: int = 0

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._start_time = datetime.datetime.now()
            cls._instance._init_store()
        return cls._instance

    def __getitem__(self, name):
        with self.lock:
            return self.data[name]

    def __iter__(self):
        with self.lock:
            return iter(self.data)

    def keys(self):
        with self.lock:
            return self.data.keys()

    def items(self):
        with self.lock:
            return self.data.items()

    def values(self):
        with self.lock:
            return self.data.values()

    def stats(self, serializable=False):
        with self.lock:
            stats = {
                "total_tracked": self.total_tracked,
            }
            pkts = {}
            for key in self.data:
                last_send_time = self.data[key].last_send_time
                pkts[key] = {
                    "last_send_time": last_send_time,
                    "send_count": self.data[key].send_count,
                    "retry_count": self.data[key].retry_count,
                    "message": self.data[key].raw,
                }
            stats["packets"] = pkts
        return stats

    def rx(self, packet: type[core.Packet]) -> None:
        """When we get a packet from the network, check if we should remove it."""
        if isinstance(packet, core.AckPacket):
            self._remove(packet.msgNo)
        elif isinstance(packet, core.RejectPacket):
            self._remove(packet.msgNo)
        elif hasattr(packet, "ackMsgNo"):
            # Got a piggyback ack, so remove the original message
            self._remove(packet.ackMsgNo)

    def tx(self, packet: type[core.Packet]) -> None:
        """Add a packet that was sent."""
        with self.lock:
            key = packet.msgNo
            packet.send_count = 0
            self.data[key] = packet
            self.total_tracked += 1

    def remove(self, key):
        self._remove(key)

    def _remove(self, key):
        with self.lock:
            try:
                del self.data[key]
            except KeyError:
                pass



import datetime
import socket
import unittest
from unittest import mock

import aprslib

from aprsd import exception
from aprsd.client.drivers.registry import ClientDriver
from aprsd.client.drivers.tcpkiss import TCPKISSDriver
from aprsd.packets import core


class TestTCPKISSDriver(unittest.TestCase):
    """Unit tests for the TCPKISSDriver class."""

    def setUp(self):
        # Mock configuration
        self.conf_patcher = mock.patch('aprsd.client.drivers.tcpkiss.CONF')
        self.mock_conf = self.conf_patcher.start()

        # Configure KISS settings
        self.mock_conf.kiss_tcp.enabled = True
        self.mock_conf.kiss_tcp.host = '127.0.0.1'
        self.mock_conf.kiss_tcp.port = 8001
        self.mock_conf.kiss_tcp.path = ['WIDE1-1', 'WIDE2-1']

        # Mock socket
        self.socket_patcher = mock.patch('aprsd.client.drivers.tcpkiss.socket')
        self.mock_socket_module = self.socket_patcher.start()
        self.mock_socket = mock.MagicMock()
        self.mock_socket_module.socket.return_value = self.mock_socket

        # Mock select
        self.select_patcher = mock.patch('aprsd.client.drivers.tcpkiss.select')
        self.mock_select = self.select_patcher.start()

        # Create an instance of the driver
        self.driver = TCPKISSDriver()

    def tearDown(self):
        self.conf_patcher.stop()
        self.socket_patcher.stop()
        self.select_patcher.stop()

    def test_implements_client_driver_protocol(self):
        """Test that TCPKISSDriver implements the ClientDriver Protocol."""
        # Verify the instance is recognized as implementing the Protocol
        self.assertIsInstance(self.driver, ClientDriver)

        # Verify all required methods are present with correct signatures
        required_methods = [
            'is_enabled',
            'is_configured',
            'is_alive',
            'close',
            'send',
            'setup_connection',
            'set_filter',
            'login_success',
            'login_failure',
            'consumer',
            'decode_packet',
            'stats',
        ]

        for method_name in required_methods:
            self.assertTrue(
                hasattr(self.driver, method_name),
                f'Missing required method: {method_name}',
            )

    def test_init(self):
        """Test initialization sets default values."""
        self.assertFalse(self.driver._connected)
        self.assertIsInstance(self.driver.keepalive, datetime.datetime)
        self.assertFalse(self.driver._running)

    def test_transport_property(self):
        """Test transport property returns correct value."""
        self.assertEqual(self.driver.transport, 'tcpkiss')

    def test_is_enabled_true(self):
        """Test is_enabled returns True when KISS TCP is enabled."""
        self.mock_conf.kiss_tcp.enabled = True
        self.assertTrue(TCPKISSDriver.is_enabled())

    def test_is_enabled_false(self):
        """Test is_enabled returns False when KISS TCP is disabled."""
        self.mock_conf.kiss_tcp.enabled = False
        self.assertFalse(TCPKISSDriver.is_enabled())

    def test_is_configured_true(self):
        """Test is_configured returns True when properly configured."""
        with mock.patch.object(TCPKISSDriver, 'is_enabled', return_value=True):
            self.mock_conf.kiss_tcp.host = '127.0.0.1'
            self.assertTrue(TCPKISSDriver.is_configured())

    def test_is_configured_false_no_host(self):
        """Test is_configured returns False when host not set."""
        with mock.patch.object(TCPKISSDriver, 'is_enabled', return_value=True):
            self.mock_conf.kiss_tcp.host = None
            with self.assertRaises(exception.MissingConfigOptionException):
                TCPKISSDriver.is_configured()

    def test_is_configured_false_not_enabled(self):
        """Test is_configured returns False when not enabled."""
        with mock.patch.object(TCPKISSDriver, 'is_enabled', return_value=False):
            self.assertFalse(TCPKISSDriver.is_configured())

    def test_is_alive(self):
        """Test is_alive property returns connection state."""
        self.driver._connected = True
        self.assertTrue(self.driver.is_alive)

        self.driver._connected = False
        self.assertFalse(self.driver.is_alive)

    def test_close(self):
        """Test close method calls stop."""
        with mock.patch.object(self.driver, 'socket') as mock_socket:
            self.driver.close()
            mock_socket.close.assert_called_once()
            self.assertFalse(self.driver._connected)

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_setup_connection_success(self, mock_log):
        """Test setup_connection successfully connects."""
        # Mock the connect method to succeed
        is_en = self.driver.is_enabled
        is_con = self.driver.is_configured
        self.driver.is_enabled = mock.MagicMock(return_value=True)
        self.driver.is_configured = mock.MagicMock(return_value=True)
        with mock.patch.object(
            self.driver, 'connect', return_value=True
        ) as mock_connect:
            self.driver.setup_connection()
            mock_connect.assert_called_once()
            mock_log.info.assert_called_with('KISS TCP Connection to 127.0.0.1:8001')

        self.driver.is_enabled = is_en
        self.driver.is_configured = is_con

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_setup_connection_failure(self, mock_log):
        """Test setup_connection handles connection failure."""
        # Mock the connect method to fail
        with mock.patch.object(
            self.driver, 'connect', return_value=False
        ) as mock_connect:
            self.driver.setup_connection()
            mock_connect.assert_called_once()
            mock_log.error.assert_called_with('Failed to connect to KISS interface')

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_setup_connection_exception(self, mock_log):
        """Test setup_connection handles exceptions."""
        # Mock the connect method to raise an exception
        with mock.patch.object(
            self.driver, 'connect', side_effect=Exception('Test error')
        ) as mock_connect:
            self.driver.setup_connection()
            mock_connect.assert_called_once()
            mock_log.error.assert_any_call('Failed to initialize KISS interface')
            mock_log.exception.assert_called_once()
            self.assertFalse(self.driver._connected)

    def test_set_filter(self):
        """Test set_filter does nothing for KISS."""
        # Just ensure it doesn't fail
        self.driver.set_filter('test/filter')

    def test_login_success_when_connected(self):
        """Test login_success returns True when connected."""
        self.driver._connected = True
        self.assertTrue(self.driver.login_success())

    def test_login_success_when_not_connected(self):
        """Test login_success returns False when not connected."""
        self.driver._connected = False
        self.assertFalse(self.driver.login_success())

    def test_login_failure(self):
        """Test login_failure returns success message."""
        self.assertEqual(self.driver.login_failure(), 'Login successful')

    @mock.patch('aprsd.client.drivers.tcpkiss.ax25frame.Frame.ui')
    def test_send_packet(self, mock_frame_ui):
        """Test sending an APRS packet."""
        # Create a mock frame
        mock_frame = mock.MagicMock()
        mock_frame_bytes = b'mock_frame_data'
        mock_frame.__bytes__ = mock.MagicMock(return_value=mock_frame_bytes)
        mock_frame_ui.return_value = mock_frame

        # Set up the driver
        self.driver.socket = self.mock_socket
        self.driver.path = ['WIDE1-1', 'WIDE2-1']

        # Create a mock packet
        mock_packet = mock.MagicMock(spec=core.Packet)
        mock_bytes = b'Test packet data'
        mock_packet.__bytes__ = mock.MagicMock(return_value=mock_bytes)
        # Add path attribute to the mock packet
        mock_packet.path = None

        # Send the packet
        self.driver.send(mock_packet)

        # Check that frame was created correctly
        mock_frame_ui.assert_called_once_with(
            destination='APZ100',
            source=mock_packet.from_call,
            path=self.driver.path,
            info=mock_packet.payload.encode('US-ASCII'),
        )

        # Check that socket send was called
        self.mock_socket.send.assert_called_once()

        # Verify packet counters updated
        self.assertEqual(self.driver.packets_sent, 1)
        self.assertIsNotNone(self.driver.last_packet_sent)

    def test_send_with_no_socket(self):
        """Test send raises exception when socket not initialized."""
        self.driver.socket = None
        mock_packet = mock.MagicMock(spec=core.Packet)

        with self.assertRaises(Exception) as context:
            self.driver.send(mock_packet)
        self.assertIn('KISS interface not initialized', str(context.exception))

    def test_stats(self):
        """Test stats method returns correct data."""
        # Set up test data
        self.driver._connected = True
        self.driver.path = ['WIDE1-1', 'WIDE2-1']
        self.driver.packets_sent = 5
        self.driver.packets_received = 3
        self.driver.last_packet_sent = datetime.datetime.now()
        self.driver.last_packet_received = datetime.datetime.now()

        # Get stats
        stats = self.driver.stats()

        # Check stats contains expected keys
        expected_keys = [
            'client',
            'transport',
            'connected',
            'path',
            'packets_sent',
            'packets_received',
            'last_packet_sent',
            'last_packet_received',
            'connection_keepalive',
            'host',
            'port',
        ]
        for key in expected_keys:
            self.assertIn(key, stats)

        # Check some specific values
        self.assertEqual(stats['client'], 'TCPKISSDriver')
        self.assertEqual(stats['transport'], 'tcpkiss')
        self.assertEqual(stats['connected'], True)
        self.assertEqual(stats['packets_sent'], 5)
        self.assertEqual(stats['packets_received'], 3)

    def test_stats_serializable(self):
        """Test stats with serializable=True converts datetime to ISO format."""
        self.driver.keepalive = datetime.datetime.now()

        stats = self.driver.stats(serializable=True)

        # Check keepalive is a string in ISO format
        self.assertIsInstance(stats['connection_keepalive'], str)
        # Try parsing it to verify it's a valid ISO format
        try:
            datetime.datetime.fromisoformat(stats['connection_keepalive'])
        except ValueError:
            self.fail('keepalive is not in valid ISO format')

    def test_connect_success(self):
        """Test successful connection."""
        result = self.driver.connect()

        self.assertTrue(result)
        self.assertTrue(self.driver._connected)
        self.mock_socket.connect.assert_called_once_with(
            (self.mock_conf.kiss_tcp.host, self.mock_conf.kiss_tcp.port)
        )
        self.mock_socket.settimeout.assert_any_call(5.0)
        self.mock_socket.settimeout.assert_any_call(0.1)

    def test_connect_failure_socket_error(self):
        """Test connection failure due to socket error."""
        self.mock_socket.connect.side_effect = socket.error('Test socket error')

        result = self.driver.connect()

        self.assertFalse(result)
        self.assertFalse(self.driver._connected)

    def test_connect_failure_timeout(self):
        """Test connection failure due to timeout."""
        self.mock_socket.connect.side_effect = socket.timeout('Test timeout')

        result = self.driver.connect()

        self.assertFalse(result)
        self.assertFalse(self.driver._connected)

    def test_fix_raw_frame(self):
        """Test fix_raw_frame removes KISS markers and handles FEND."""
        # Create a test frame with KISS markers
        with mock.patch(
            'aprsd.client.drivers.tcpkiss.handle_fend', return_value=b'fixed_frame'
        ) as mock_handle_fend:
            raw_frame = b'\xc0\x00some_frame_data\xc0'  # \xc0 is FEND

            result = self.driver.fix_raw_frame(raw_frame)

            mock_handle_fend.assert_called_once_with(b'some_frame_data')
            self.assertEqual(result, b'fixed_frame')

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_decode_packet_success(self, mock_log):
        """Test successful packet decoding."""
        mock_frame = 'test frame data'
        mock_aprs_data = {'from': 'TEST-1', 'to': 'APRS'}
        mock_packet = mock.MagicMock(spec=core.Packet)

        with mock.patch(
            'aprsd.client.drivers.tcpkiss.aprslib.parse', return_value=mock_aprs_data
        ) as mock_parse:
            with mock.patch(
                'aprsd.client.drivers.tcpkiss.core.factory', return_value=mock_packet
            ) as mock_factory:
                result = self.driver.decode_packet(frame=mock_frame)

                mock_parse.assert_called_once_with(str(mock_frame))
                mock_factory.assert_called_once_with(mock_aprs_data)
                self.assertEqual(result, mock_packet)

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_decode_packet_no_frame(self, mock_log):
        """Test decode_packet with no frame returns None."""
        result = self.driver.decode_packet()

        self.assertIsNone(result)
        mock_log.warning.assert_called_once()

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_decode_packet_exception(self, mock_log):
        """Test decode_packet handles exceptions."""
        mock_frame = 'invalid frame'

        with mock.patch(
            'aprsd.client.drivers.tcpkiss.aprslib.parse',
            side_effect=Exception('Test error'),
        ) as mock_parse:
            result = self.driver.decode_packet(frame=mock_frame)

            mock_parse.assert_called_once()
            self.assertIsNone(result)
            mock_log.error.assert_called_once()

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_consumer_with_frame(self, mock_log):
        """Test consumer processes frames and calls callback."""
        mock_callback = mock.MagicMock()
        mock_frame = mock.MagicMock()

        # Configure driver for test
        self.driver._connected = True
        self.driver._running = True

        # Set up read_frame to return one frame then stop
        def side_effect():
            self.driver._running = False
            return mock_frame

        with mock.patch.object(
            self.driver, 'read_frame', side_effect=side_effect
        ) as mock_read_frame:
            self.driver.consumer(mock_callback)

            mock_read_frame.assert_called_once()
            mock_callback.assert_called_once_with(frame=mock_frame)

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_read_frame_success(self, mock_log):
        """Test read_frame successfully reads a frame."""
        # Set up driver
        self.driver.socket = self.mock_socket
        self.driver._connected = True

        # Mock socket recv to return data
        raw_data = b'\xc0\x00test_frame\xc0'
        self.mock_socket.recv.return_value = raw_data

        # Mock select to indicate socket is readable
        self.mock_select.select.return_value = ([self.mock_socket], [], [])

        # Mock fix_raw_frame and Frame.from_bytes
        mock_fixed_frame = b'fixed_frame'
        mock_ax25_frame = mock.MagicMock()

        with mock.patch.object(
            self.driver, 'fix_raw_frame', return_value=mock_fixed_frame
        ) as mock_fix:
            with mock.patch(
                'aprsd.client.drivers.tcpkiss.ax25frame.Frame.from_bytes',
                return_value=mock_ax25_frame,
            ) as mock_from_bytes:
                result = self.driver.read_frame()

                self.mock_socket.setblocking.assert_called_once_with(0)
                self.mock_select.select.assert_called_once()
                self.mock_socket.recv.assert_called_once()
                mock_fix.assert_called_once_with(raw_data)
                mock_from_bytes.assert_called_once_with(mock_fixed_frame)
                self.assertEqual(result, mock_ax25_frame)

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_read_frame_select_timeout(self, mock_log):
        """Test read_frame handles select timeout."""
        # Set up driver
        self.driver.socket = self.mock_socket
        self.driver._running = True

        # Mock select to indicate no readable sockets
        self.mock_select.select.return_value = ([], [], [])

        result = self.driver.read_frame()

        self.assertIsNone(result)

    @mock.patch('aprsd.client.drivers.tcpkiss.LOG')
    def test_read_frame_socket_error(self, mock_log):
        """Test read_frame handles socket error."""
        # Set up driver
        self.driver.socket = self.mock_socket
        self.driver._connected = True

        # Mock setblocking to raise OSError
        self.mock_socket.setblocking.side_effect = OSError('Test error')

        with self.assertRaises(aprslib.ConnectionDrop):
            self.driver.read_frame()
            mock_log.error.assert_called_once()


if __name__ == '__main__':
    unittest.main()
